package no.skippy.agent;

import no.skippy.agent.event.AfterTurnEvent;
import no.skippy.agent.event.RoundEndedEvent;
import no.skippy.agent.event.TurnEvent;
import no.skippy.agent.util.Storage;
import no.skippy.sim.control.SimulatorTask;

/**
 * A event interface for receiving basic agent events. * @author <a
 * href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public interface Agent extends InterActive {

	/**
	 * Agent receive after the agent's rock is sent. (Before the other player
	 * has sent its rock).
	 * <p/>
	 * Order the event is sent by the {@link SimulatorTask}.
	 * <ul>
	 * <li>agent1.onTurn(TurnEvent)</li>
	 * <li>agent1.afterTurn(AfterTurnEvent)</li>
	 * <li>agent2.onTurn(TurnEvent)</li>
	 * </ul>
	 * <p/>
	 * By default {@link Agent#onTurn(TurnEvent)} does nothing.
	 */
	void afterTurn(AfterTurnEvent event);

	/**
	 * 
	 * @return the name of the agent.
	 */
	String getAgentName();

	/**
	 * Agent receive if it lost the round.
	 * @param event for a round.
	 */
	void onLoose(RoundEndedEvent event);

	/**
	 * Agent receive if the round ended with tie.
	 * <p/>
	 * By default {@link Agent#onTie(RoundEndedEvent)} does nothing.
	 * @param event for a round.
	 */
	void onTie(RoundEndedEvent event);

	/**
	 * Agent receive when it is its turn.
	 * <p/>
	 * By default {@link Agent#onTurn(TurnEvent)} does nothing.
	 */
	void onTurn(TurnEvent event);

	/**
	 * Agent receive if it won the round.
	 * @param event for a round.
	 */
	void onWin(RoundEndedEvent event);
	
	/**
	 * Load agent resources. Occurs before {@link Agent#setup()}.
	 * @param service to use.
	 */
	void loadResources(Storage service);
	
	/**
	 * Get loaded resource.
	 * @param key to use.
	 * @return resource related with the key.
	 */
	Object getResource(String key);

	/**
	 * Responsible for initialisations that must occur before the {@link Agent}
	 * can play curling.
	 * <p/>
	 * By default {@link Agent#setup()} does nothing.
	 */
	void setup();

	/**
	 * Called after the simulator {@link SimulatorTask} has completed a curling
	 * match. </p> Subclasses may override this method to do any cleanup tasks
	 * that are necessary before exiting. For example save what the agent has
	 * learned.
	 * @param service to use.
	 */
	void tearDown(Storage service);
	
	void addResource(String key, Object value);
}
