package no.skippy.agent.event;

import java.awt.geom.Point2D;
import java.util.Set;

import javax.swing.BoundedRangeModel;

import no.skippy.agent.Agent;
import no.skippy.sim.control.BroomControls;
import no.skippy.sim.control.ReadOnlyRockCollection;

/**
 * Event type passed to {@link Agent} when it is its turn.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class TurnEvent extends Event implements BroomControls {

	/** */
	private static final long serialVersionUID = 1L;

	/** Broom controller for the agent. */
	private BroomControls iAgentControls = null;

	/** all rock positions, include rocks not yet played. */
	private ReadOnlyRockCollection rockCollection;

	/** set of rocks current in play. (not out). */
	private Set<Integer> rocksInPlay;

	/** is rock/player dark. */
	private boolean isDark = false;

	/** rock to set. */
	private int rock;

	/** playing in round. */
	private int round;

	/** Total expected number of rounds. */
	private int numberOfRounds;

	/** has last rock. */
	private boolean hasHammer;
	
	/** the opponent agent name.*/
	private String opponent;

	/**
	 * Constructor.
	 * 
	 * @param broomController
	 *            broom agent controller.
	 * @param rockCollection
	 *            of all rocks, including rocks not yet played.
	 * @param rocksInPlay
	 *            set of rocks played, and still in game.
	 * @param isDark
	 *            agent rock colour. (Dark is red).
	 * @param rockidx16
	 *            playing.
	 * @param round
	 *            playing.
	 * @param numberOfRounds
	 *            total number of rounds.
	 * @param hasHammer
	 *            has last rock in round.
	 */
	public TurnEvent(BroomControls broomController,
			ReadOnlyRockCollection rockCollection, Set<Integer> rocksInPlay,
			boolean isDark, int rockidx16, int round, int numberOfRounds,
			boolean hasHammer, String opponent) {
		this.iAgentControls = broomController;
		this.rockCollection = rockCollection;
		this.rocksInPlay = rocksInPlay;
		this.isDark = isDark;
		this.rock = rockidx16;
		this.round = round;
		this.numberOfRounds = numberOfRounds;
		this.hasHammer = hasHammer;
		this.opponent = opponent;
	}

	/**
	 * Get current rock position.
	 * 
	 * @param i16
	 *            rock id.
	 * @return the position of the rock.
	 */
	public Point2D getCurrentPos(int i16) {
		return rockCollection.getCurrentPos(i16);

	}

	/**
	 * Get position of a rock's positions before setting the rock.
	 * 
	 * @param i16
	 *            rock id.
	 * @return the position of the rock.
	 */
	public Point2D getInitialPos(int i16) {
		return rockCollection.getInitialPos(i16);
	}

	public Point2D getBroom() {
		return iAgentControls.getBroom();
	}

	public boolean getOutTurn() {
		return iAgentControls.getOutTurn();
	}

	/**
	 * Rock is dark.
	 * 
	 * @param i16
	 *            rock id.
	 * @return if the rock is dark (red).
	 */
	public boolean isRockDark(int i16) {
		return i16 % 2 == 0;
	}

	public BoundedRangeModel getSplitTimeMillies() {
		return iAgentControls.getSplitTimeMillies();
	}

	public void setBroom(Point2D pos) {
		iAgentControls.setBroom(pos);
	}

	public void setOutTurn(boolean outTurn) {
		iAgentControls.setOutTurn(outTurn);
	}

	public void setSplitTimeMillies(int splitTime) {
		iAgentControls.setSplitTimeMillies(splitTime);
	}

	/**
	 * @return a set of rocks that is in play.
	 */
	public Set<Integer> getRocksInPlay() {
		return rocksInPlay;
	}

	/**
	 * @return true if the player is dark (red).
	 */
	public boolean isPlayerDark() {
		return isDark;
	}

	/**
	 * @return the rock id controlled by the broom.
	 */
	public int getRock() {
		return rock;
	}

	/**
	 * @return the current curling round.
	 */
	public int getRound() {
		return round;
	}

	/**
	 * @return the number of rounds to play. Default curling rules dictates 10
	 *         rounds.
	 */
	public int getNumberOfRounds() {
		return numberOfRounds;
	}

	/**
	 * @return collection of rock positions. (read only)
	 */
	public ReadOnlyRockCollection getAllRocks() {
		return rockCollection;
	}

	/**
	 * @return true if the event receiver (agent) has last rock in the current
	 *         round.
	 */
	public boolean isHasHammer() {
		return hasHammer;
	}

	/** 
	 * @return the opponents agent name.
	 */
	public String getOponent() {
		return opponent;
	}
	
	/**
	 * @param rock
	 *            i16 rock id.
	 * @return true if the rock belongs to the agent.
	 */
	public boolean isMyRock(int rock) {
		return isRockDark(rock) && isPlayerDark()
				|| (!isRockDark(rock) && !isPlayerDark());
	}
}
