package no.skippy.agent.examples;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;

/**
 * Class for handling a linear target function.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class LFunction {

	private String[] description;
	private int[] features;
	private double[] weights;
	private double constant = .001;
	private int trainingCount = 0;
	private int roundsPlayed = 0;

	public LFunction(int numOfFeatures, double minRand, double maxRand) {
		features = new int[numOfFeatures];
		weights = new double[numOfFeatures + 1];

		for (int i = 0; i < weights.length; i++) { // random weights
			weights[i] = (Math.random() * (maxRand - minRand + 1)) + minRand;
		}
	}

	Log log = JCLoggerFactory.getLogger(LFunction.class);
	
	/**
	 * Calculate a estimate of the target function V.
	 * 
	 * @param features
	 *            the linear features (xi's).
	 * @return estimate of V.
	 */
	public double getEstimateV(int[] features) {

		double vEstimate = 0.0;
		vEstimate += weights[0]; // w0
		for (int i = 0; i < features.length; i++) {
			vEstimate += weights[i + 1] * features[i];
		}

		// setFeatures(xi);
		return vEstimate;
	}

	/**
	 * Update the weights.
	 * 
	 * @param V
	 *            value.
	 * @return updated weights.
	 */
	public double[] updateWeights(double V, double constant) {
		double estimatedV = getEstimateV(features);
		log.info("EstimatedV: " + estimatedV + "\tRealV:" + V);
		
		for (int i = 1; i < weights.length; i++) {
			weights[i] = weights[i] + (constant * (V - estimatedV))
					* features[i - 1];
		}
		return weights;
	}

	/**
	 * @return features
	 */
	public int[] getFeatures() {
		return features;
	}

	/**
	 * @param features
	 */
	public void setFeatures(int[] features) {
		this.features = features;
	}

	/**
	 * @return feature weights.
	 */
	public double[] getWeights() {
		return weights;
	}

	/**
	 * @param weights
	 *            to set.
	 */
	public void setWeights(double[] weights) {
		this.weights = weights;
	}

	public double getConstant() {
		return constant;
	}

	public void setConstant(double constant) {
		this.constant = constant;
	}

	public int getTrainingCount() {
		return trainingCount;
	}

	public void setTrainingCount(int trainingCount) {
		this.trainingCount = trainingCount;
	}

	public int getRoundsPlayed() {
		return roundsPlayed;
	}

	public void setRoundsPlayed(int roundsPlayed) {
		this.roundsPlayed = roundsPlayed;
	}

	public String[] getDescription() {
		return description;
	}

	public void setDescription(String[] description) {
		this.description = description;
	}

	/**
	 * Clones the object, creates a new LFunction with features and weight
	 * values.
	 */
	public LFunction clone() {
		LFunction lf = new LFunction(features.length, 1, 1);
		lf.setWeights(weights.clone()); // copy over the weights
		lf.setFeatures(features.clone());
		lf.setDescription(description);
		// ignores the features, not necessary.
		return lf;
	}
}
