package no.skippy.agent.skippy;

import java.io.File;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import no.skippy.agent.Agent;
import no.skippy.agent.BasicAgent;
import no.skippy.agent.event.TurnEvent;
import no.skippy.agent.examples.LFunction;
import no.skippy.agent.examples.LFunctionLog;
import no.skippy.agent.skippy.api.Action;
import no.skippy.agent.skippy.api.Featurizer;
import no.skippy.agent.skippy.api.State;
import no.skippy.agent.util.GameLog;
import no.skippy.agent.util.Storage;

/**
 * The base class for Skippy Agents. </p> This class defines a collection of
 * useful methods for Skippy agents.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public abstract class AbstractAgent extends BasicAgent {

	public static SimpleDateFormat DATE_STAMP = new SimpleDateFormat(
			"dd.MM.yy_hhmm");

	/** name of the agent, @see {@link Agent#getAgentName()}*/
	private String agentName;
	
	/** suffix to agent weight XML file. */
	private String AGENT_DATA;

	/** suffix to agent weight XML file. */
	private String AGENT_DATA_HAMMER;
	
	/** root folder for agent logs.*/
	private String folderPath = "agentlog/" + getClass().getSimpleName() + "/"
			+ getClass().getSimpleName();

	/** Path to where agent writes its weight log. */
	protected String LOG_FILE = folderPath + "_%s";

	/** Path to where agent writes its hammer weight log. */
	protected String LOG_FILE_HAMMER = folderPath + "_Hammer_%s";

	/** Log of the weights when not having hammer. New entry is added after each training. */
	protected GameLog<LFunctionLog> logNormal = new GameLog<LFunctionLog>();
	
	/** Log of the weights when having hammer. New entry is added after each training. */
	protected GameLog<LFunctionLog> logHammer = new GameLog<LFunctionLog>();

	/** Log of the state transitions. */
	protected List<State> stateLog = new ArrayList<State>();

	/** Entry name for the {@link GameLog} log variable. */
	protected String logEntry = getClass().getSimpleName();

	/** Total number of rounds played. */
	protected int roundsPlayed = 0;

	/** Best estimated action, */
	protected Action bestAction;

	/** State result of executing best action. */
	protected State bestState;

	/** True if agent should use separate hammer weights.*/
	private boolean useHammer = false;
	
	/** Count of time training normal features. */
	private int trainingCountNormal = 0;
	
	/** Count of times training hammer features.*/
	private int trainingCountHammer = 0;
	
	/** Weight used in a hammer round. */
	protected Featurizer hammerFeatures;

	/** Weights used in a normal round. */
	protected Featurizer normalFeatures;
	
	/** Current Weights used to estimate value of a state. */
	protected Featurizer features;

	/** True if the agent has hammer in the round.*/
	protected boolean hammerRound = false;
	
	/**True if the agent should write logs to disk after training.*/
	private boolean writeOut = true;

	/**
	 * Configure the Agent.
	 * 
	 * <ul>
	 * <li>Agent Name {@link #agentName}</li>
	 * <li>Use one set of weights {@link #useHammer}</li>
	 * <li>Agent Data weight prefix {@link #AGENT_DATA}</li>
	 * <li>Agent Data weight prefix {@link #AGENT_DATA_HAMMER}</li>
	 * </ul>
	 * 
	 */
	abstract protected void configure();
	
	public AbstractAgent() {
		configure();
	}
	
	@Override
	public void loadResources(Storage service) {
		service.loadObject(LFunction.class, getAgentData(), this);
		if (isUseHammer())
			service.loadObject(LFunction.class, getAgentDataHammer(), this);
	}

	public void tearDown(Storage service) {
		if (!writeOut) return;
		String date = DATE_STAMP.format(new Date());
		String filePath = String.format(LOG_FILE + JSKIP, date);

		// write weight log
		service.saveResource(logNormal, filePath = getFilePath(filePath));
		log(WROTE_TOFILE, getClass().getSimpleName(), filePath);

		// prepare agent data
		LFunction lf = prepareAgentData(normalFeatures);
		filePath = String.format(LOG_FILE + XML, date);
		service.saveResource(lf, filePath = getFilePath(filePath));
		log(WROTE_TOFILE, getClass().getSimpleName(), filePath);

		if (isUseHammer()) {
			lf = prepareAgentData(hammerFeatures);
			filePath = String.format(LOG_FILE_HAMMER + XML, date);
			service.saveResource(lf, filePath = getFilePath(filePath));
			log(WROTE_TOFILE, getClass().getSimpleName(), filePath);
			
			filePath = String.format(LOG_FILE_HAMMER + JSKIP, date);
			
			service.saveResource(logHammer, filePath = getFilePath(filePath));
			log(WROTE_TOFILE, getClass().getSimpleName(), filePath);
		}
	}

	public String getAgentName() {
		return agentName;
	}
	
	
	/**
	 * Get new state from using an action on a state.
	 * 
	 * @param action
	 *            to execute
	 * @param event
	 *            to use.
	 * @return result {@link State} from an action.
	 */
	protected State getState(Action action, TurnEvent event) {
		if (action != null) {
			action.execute(event);
		}
		return new AgentState(event);
	}

	/**
	 * Get state from an event.
	 * 
	 * @param event
	 *            to use.
	 * @return the state of the {@link TurnEvent}.
	 */
	public State getState(TurnEvent event) {
		return getState(null, event);
	}

	/**
	 * Train the weights.
	 * 
	 * @param points
	 *            received.
	 * @param realValue
	 *            reward.
	 */
	protected void trainWeights(int points, double realValue) {
		log(TRAIN, stateLog.size());

		if (features.getLearningRate() == 0.0) {
			log(SKIPP_TRAINING, features.getLearningRate());

		} else {
			String message = "\n\n";
			message += format(FORMAT, "", "Estimated", "Real Value");

			double[] weights = features.getFeatureWeights();
			double learningFactor = 1;

			for (int index = stateLog.size() - 1; index >= 0; index--) {
				State state = stateLog.get(index);
				double estimate = features.getEstimate(state);
				for (int i = 1; i < weights.length; i++) {
					weights[i] = features.getFeatureWeights()[i]
							+ (features.getLearningRate() / learningFactor * (realValue - estimate))
							* features.getFeatures(state)[i - 1];
				}
				features.setFeatureWeights(weights);

				learningFactor += .5;

				message += format(FORMAT, "", DIGIT.format(estimate), DIGIT
						.format(realValue));
			}

			incTrainingCount(features);
			stateLog.clear();

			double trainingCountDouble = getTrainingCount(features);
			features.setLearningRate(1 / (200 + trainingCountDouble));

			log(message);
			log(AFTER_TRAINING, getTrainingCount(features), roundsPlayed);
		}

		printFeatureWeights(features, bestState, hammerRound);
		addLog(points, features.getEstimate(bestState), realValue);
	}

	/**
	 * Log training.
	 * 
	 * @param rockPoints
	 *            in the round.
	 * @param estimatedValue
	 *            estimated value of the round.
	 * @param realValue
	 *            reward value of the round.
	 */
	protected void addLog(int rockPoints, double estimatedValue,
			double realValue) {
		GameLog<LFunctionLog> log = hammerRound?logHammer:logNormal;
		log.addLog(logEntry, new LFunctionLog(features.getFeatureWeights(),
				features.getLearningRate(), rockPoints, roundsPlayed,
				estimatedValue, realValue));
	}

	/**
	 * Log state
	 * 
	 * @param state
	 *            to log.
	 */
	protected void addStateLog(State state) {
		stateLog.add(state);
	}

	/**
	 * Print the feature weights.
	 * 
	 * @param features
	 *            weights to print.
	 * @param state
	 *            , used to get weights description.
	 */
	protected void printFeatureWeights(Featurizer features, State state,
			boolean hammer) {
		String weightLog = "\n\nFeature Weights: (rounded) "
				+ (hammer ? " Hammer" : "Normal") + "\n";
		weightLog += format(FORMAT, "Weight", "Value", "Description");
		for (int j = 0; j < features.getFeatureWeights().length; j++) {
			weightLog += String.format(FORMAT, new Object[] { "w" + j + ":",
					DIGIT.format(features.getFeatureWeights()[j]),
					features.getFeatureWeightsDescription(state)[j] });
		}
		weightLog += format(FORMAT, "Learningrate: ", features
				.getLearningRate(), "");

		log(weightLog);
	}

	/**
	 * Get a unused filename. If the filePath already exist, '.new' is appended.
	 * 
	 * @param filePath
	 *            to use.
	 * @return unused filename.
	 */
	public String getFilePath(String filePath) {
		while (new File(filePath).exists()) {
			log(FILE_EXIST, filePath);
			filePath += ".new";
		}
		return filePath;
	}

	/**
	 * Get formatted string.
	 * 
	 * @param format
	 *            to use.
	 * @param args
	 *            to include.
	 * @return formatted string.
	 */
	protected String format(String format, Object... args) {
		return String.format(format, args);
	}

	private LFunction prepareAgentData(Featurizer featurizer) {
		LFunction lf = new LFunction(AgentState.NUM_FEATURES, 0, 0);
		lf.setConstant(featurizer.getLearningRate());
		lf.setDescription(featurizer.getFeatureWeightsDescription(bestState));
		lf.setRoundsPlayed(roundsPlayed);
		lf.setTrainingCount(getTrainingCount(featurizer));
		lf.setWeights(featurizer.getFeatureWeights());
		return lf;
	}

	protected String getAgentData() {
		return folderPath + "_" + this.AGENT_DATA + XML;
	}
	
	protected String getAgentDataHammer() {
		return folderPath + "_Hammer_" + AGENT_DATA_HAMMER + XML;
	}
	
	protected void incTrainingCount(Featurizer features) {
		if (features == normalFeatures) {
			trainingCountNormal += 1;
			return;
		} else if (features == hammerFeatures) {
			trainingCountHammer += 1;
			return;
		}
		throw new IllegalStateException("Unknown featurizer used");
	}
	
	protected void setTrainingCount(Featurizer features, int trainingCount) {
		if (features == normalFeatures) {
			trainingCountNormal = trainingCount;
			return;
		} else if (features == hammerFeatures) {
			trainingCountHammer = trainingCount;
			return;
		}
		throw new IllegalStateException("Unknown featurizer used");
	}
	
	protected int getTrainingCount(Featurizer features) {
		if (features == normalFeatures) {
			return trainingCountNormal;
		} else if (features == hammerFeatures) {
			return trainingCountHammer;
		}
		throw new IllegalStateException("Unknown featurizer used");
	}
	
	protected void useHammer(boolean useHammer) {
		this.useHammer = useHammer;
	}
	
	protected boolean isUseHammer() {
		return useHammer;
	}
	
	protected void setAgentName(String agentName) {
		this.agentName = agentName; 
	}
	
	protected void setNormalSuffix(String suffix) {
		AGENT_DATA = suffix;
	}

	protected void setHammerSuffix(String suffix) {
		AGENT_DATA_HAMMER = suffix;
	}
	
	public void setWriteOut(boolean writeOut) {
		this.writeOut = writeOut;
	}


	protected static DecimalFormat DIGIT = new DecimalFormat("#,##0.000");

	protected static String FORMAT = "%1$-10s%2$-10s\t%3$-20s\n";

	protected static String FILE_EXIST = "FilePath: %s allready exist.";

	protected static String WROTE_TOFILE = "Wrote %s data to: %s.";

	protected static String AFTER_TRAINING = "Weights after training: %s Round: %s";

	protected static String TRAIN = "Training using experience from %s rounds.";

	protected static String SKIPP_TRAINING = "Did not train, since learning rate was %s.";

	protected static String XML = ".xml";

	protected String JSKIP = ".jskip";
}
