package no.skippy.agent.skippy;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import no.skippy.agent.event.TurnEvent;
import no.skippy.agent.skippy.api.Action;

/**
 * Action implementation used by {@link SkippyAgent}. Changes the properties of
 * broom.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class AgentAction implements Action {

	/** List of sub-actions.*/
	protected List<Action> subActions = new ArrayList<Action>();

	/** Position of the broom. */
	private Point2D pos;

	/** curl out */
	private boolean outTurn;

	/** Rock speed. */
	private int splitTime;

	/** Default minimum x position of the broom. */
	public static double DEFAULT_MIN_X = -2.0;
	/** Default maximum x position of the broom. */
	public static double DEFAULT_MAX_X = 2.0;
	/** Default step/increment of x for the broom. */
	public static double DEFAULT_STEP_X = 0.5;
	/** Default minimum split time of the broom. */
	public static double DEFAULT_MIN_SPLIT = 3000;
	/** Default maximum split time of the broom. */
	public static double DEFAULT_MAX_SPLIT = 3700;
	/** Default step/increment of split time for the broom. */
	public static double DEFAULT_STEP_SPLIT = 100;

	protected double minX, maxX, stepX;
	protected double minSplit, maxSplit, stepSplit;

	/**
	 * Create a basic action, using the broom properties available in the
	 * {@link TurnEvent} argument. Creates the list of sub-actions using the
	 * Default values.
	 * 
	 * @param event
	 *            broom properties to use.
	 */
	public AgentAction(TurnEvent event) {
		this(event, true);
	}

	/**
	 * Create a basic action, with a list of sub-actions using the Default
	 * values.
	 * 
	 * @param createSubActions
	 *            true if sub-actions should be created.
	 */
	public AgentAction(boolean createSubActions) {
		pos = new Point2D.Double(0.0, 0.0);
		outTurn = false;

		minX = DEFAULT_MIN_X;
		maxX = DEFAULT_MAX_X;
		stepX = DEFAULT_STEP_X;

		minSplit = DEFAULT_MIN_SPLIT;
		maxSplit = DEFAULT_MAX_SPLIT;
		stepSplit = DEFAULT_STEP_SPLIT;

		if (createSubActions) {
			createSubActions(true);
		}
	}

	/**
	 * Create a basic action, using the broom properties available in the
	 * {@link TurnEvent} argument. Creates the list of sub-actions using the
	 * Default values if createSubactions is true.
	 * 
	 * @param event
	 *            broom properties to use.
	 * @param createSubactions
	 *            true if sub-actions with default values should be created.
	 */
	public AgentAction(TurnEvent event, boolean createSubactions) {
		pos = event.getBroom();
		outTurn = event.getOutTurn();
		splitTime = event.getSplitTimeMillies().getValue();

		minX = DEFAULT_MIN_X;
		maxX = DEFAULT_MAX_X;
		stepX = DEFAULT_STEP_X;

		minSplit = DEFAULT_MIN_SPLIT;
		maxSplit = DEFAULT_MAX_SPLIT;
		stepSplit = DEFAULT_STEP_SPLIT;

		if (createSubactions) {
			createSubActions(true);
		}
	}

	/**
	 * Create a basic action, using the broom properties available in the
	 * {@link TurnEvent} argument. The rest of the arguments are used to create
	 * the sub-actions that is made available through
	 * {@link AgentAction#getSubActions()}.
	 * 
	 * @param event
	 *            broom properties to use.
	 * @param minX
	 *            minimum broom x position.
	 * @param maxX
	 *            maximum broom x position.
	 * @param stepX
	 *            increment of x position.
	 * @param minSplit
	 *            minimum broom split time.
	 * @param maxSplit
	 *            maximum broom split time.
	 * @param stepSplit
	 *            increment of split time.
	 * @param outInTurn
	 *            true if sub actions should contain both out turn and not out
	 *            turn actions. if false, only {@link AgentAction#isOutTurn()}
	 *            actions is created.
	 */
	public AgentAction(TurnEvent event, double minX, double maxX, double stepX,
			double minSplit, double maxSplit, double stepSplit,
			boolean outInTurn) {
		this(event, false);

		this.minX = minX;
		this.maxX = maxX;
		this.stepX = stepX;

		this.minSplit = minSplit;
		this.maxSplit = maxSplit;
		this.stepSplit = stepSplit;

		createSubActions(outInTurn); // create sub-actions with both curling
		// directions.
	}

	/**
	 * Create a new broom action.
	 * 
	 * @param pos
	 *            x property for the broom.
	 * @param outTurn
	 *            property for the broom.
	 * @param splitTime
	 *            property for the broom.
	 */
	public AgentAction(Point2D pos, int splitTime, boolean outTurn) {
		super();
		this.pos = pos;
		this.splitTime = splitTime;
		this.outTurn = outTurn;
	}

	/**
	 * 
	 * @param action
	 *            broom properties to use.
	 * @param minX
	 *            minimum broom x position.
	 * @param maxX
	 *            maximum broom x position.
	 * @param stepX
	 *            increment of x position.
	 * @param minSplit
	 *            minimum broom split time.
	 * @param maxSplit
	 *            maximum broom split time.
	 * @param stepSplit
	 *            increment of split time.
	 * @param outInTurn
	 *            true if sub actions should contain both out turn and not out
	 *            turn actions. if false, only {@link AgentAction#isOutTurn()}
	 *            actions is created.
	 */
	public AgentAction(Action action, double minX, double maxX, double stepX,
			double minSplit, double maxSplit, double stepSplit,
			boolean outInTurn) {

		this.minX = minX;
		this.maxX = maxX;
		this.stepX = stepX;

		this.minSplit = minSplit;
		this.maxSplit = maxSplit;
		this.stepSplit = stepSplit;

		this.pos = action.getPos();
		this.splitTime = action.getSplitTime();
		this.outTurn = action.isOutTurn();

		createSubActions(outInTurn); // create sub-actions with both curling
		// directions.
	}

	/**
	 * Clear the list of sub-actions, and create new using the boundaries from
	 * the arguments.
	 * 
	 * @param minX
	 *            minimum broom x position.
	 * @param maxX
	 *            maximum broom x position.
	 * @param stepX
	 *            increment of x position.
	 * @param minSplit
	 *            minimum broom split time.
	 * @param maxSplit
	 *            maximum broom split time.
	 * @param stepSplit
	 *            increment of split time.
	 * @param outInTurn
	 *            true if sub actions should contain both out turn and not out
	 *            turn actions. if false, only {@link AgentAction#isOutTurn()}
	 *            actions is created.
	 */
	public void clearCreateNewSubActions(final double minX, final double maxX,
			final double stepX, final double minSplit, final double maxSplit,
			final double stepSplit, final boolean outInTurn) {
		subActions.clear();

		for (double x = minX; x <= maxX; x += stepX) {
			for (double splitTime = minSplit; splitTime < maxSplit; splitTime += stepSplit) {
				subActions.add(new AgentAction(
						new Point2D.Double(x, pos.getY()), (int) splitTime,
						isOutTurn()));

				if (outInTurn) { // create both curl directions.
					subActions.add(new AgentAction(new Point2D.Double(x, pos
							.getY()), (int) splitTime, !isOutTurn()));
				}
			}
		}
	}

	public Point2D getPos() {
		return pos;
	}

	public boolean isOutTurn() {
		return outTurn;
	}

	public int getSplitTime() {
		return splitTime;
	}

	public void execute(TurnEvent event) {
		event.setBroom(getPos());
		event.setOutTurn(isOutTurn());
		event.setSplitTimeMillies(getSplitTime());
	}

	public List<Action> getSubActions() {
		return subActions;
	}

	public void setAction(Action action) {
		this.pos = action.getPos();
		this.splitTime = action.getSplitTime();
		this.outTurn = action.isOutTurn();
	}

	/**
	 * Create sub-actions.
	 * 
	 * @param outTurn
	 *            create actions with both curling directions.
	 */
	private void createSubActions(boolean outTurn) {
		clearCreateNewSubActions(minX, maxX, stepX, minSplit, maxSplit,
				stepSplit, outTurn);
	}

	public String getDescription() {
		return "Large Area";
	}

	public boolean equals(Object o) {
		if (o == null || !(o instanceof AgentAction))
			return false;
		AgentAction a = (AgentAction) o;
		if (getSubActions().size() != a.getSubActions().size())
			return false;
		boolean subs = true;
		for (int i = 0; i < subActions.size(); i++)
			subs &= subActions.get(i).equals(a.getSubActions().get(i));
		return a.getPos().getX() == getPos().getX()
				&& a.getPos().getY() == getPos().getY()
				&& a.getSplitTime() == getSplitTime()
				&& a.isOutTurn() == isOutTurn() && subs;
	}

	public int hashCode() {
		int hash = 7;
		hash = (int) (31 * hash + minX);
		hash = (int) (31 * hash + maxX);
		hash = (int) (31 * hash + stepX);
		hash = (int) (31 * hash + minSplit);
		hash = (int) (31 * hash + maxSplit);
		hash = (int) (31 * hash + stepSplit);
		
		hash = (int) (31 * hash + (null == pos ? 0 : pos.getX()));
		hash = (int) (31 * hash + (null == pos ? 0 : pos.getX()));
		hash = (int) (31 * hash + (outTurn? 0 : 1));
		hash = (int) (31 * hash + splitTime);
		for (Action a : subActions)
			hash = 31*hash + a.hashCode();
		return hash;
	}
}
