package no.skippy.agent.skippy;

import java.util.Arrays;

import no.skippy.agent.event.TurnEvent;
import no.skippy.agent.skippy.api.State;
import no.skippy.agent.util.CurlingUtils;

/**
 * Feature based classification of curling.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class AgentState implements State, Cloneable {

	/** Match Points. */
	protected int points;

	/** Number of Rocks that is mine and in house. */
	protected double myRocksInHouse;
	
	/** Number of Rocks that is NOT mine and in house. */
	protected double notmyRocksInHouse;

	/** Number of Rocks that is mine and covered. */
	protected int myRocksCovered;

	/** Number of Rocks that is NOT mine and covered. */
	protected int notmyRocksCovered;

	/** Number. of Corner guards. */
	protected int cornerGuard;

	/** Number of Center guards. */
	protected int centerGuard;

	/** Estimated value of the state. */
	protected double estimate;

	/** The total number of FEATURES. */
	public static final int NUM_FEATURES = 7;

	/** Default lower bound for initialising random weights. */
	public static final int LOWER_BOUNDS = -10;

	/** Default upper bound for initialising random weights. */
	public static final int UPPER_BUUNDS = 10;

	private static double IN_HOUSE_CONSTANT = CurlingUtils.feetToMeter(12);

	protected int discreteMyRocksInHouse;
	
	protected int discreteNotMyRocksInHouse;
	
	public AgentState() {
		this(0, 0, 0, 0, 0, 0, 0);
	}
	
	/**
	 * Creates a state from the inputs given by the simulator.
	 * 
	 * @param event
	 *            to use.
	 */
	public AgentState(TurnEvent event) {

		for (int rockId : event.getRocksInPlay()) {
			if (CurlingUtils.isInHouse(event.getCurrentPos(rockId))) {

				if (event.isMyRock(rockId)) {
					myRocksInHouse += inHouseScore(event.getCurrentPos(rockId)
							.distance(0, 0));
					discreteMyRocksInHouse++;
				} else {
					notmyRocksInHouse += inHouseScore(event.getCurrentPos(
							rockId).distance(0, 0));
					discreteNotMyRocksInHouse++;
				}

				for (int guardId : event.getRocksInPlay()) {
					if (guardId != rockId) {
						if ((Math.abs(event.getCurrentPos(rockId).getX()
								- event.getCurrentPos(guardId).getX()) < CurlingUtils.ROCK_RAD)
								&& event.getCurrentPos(rockId).getY() < event
										.getCurrentPos(guardId).getY()
								&& event.getCurrentPos(guardId).getY() > CurlingUtils.ROCK_RAD) {

							if (event.isMyRock(rockId)) {
								myRocksCovered++; // features[3]++;
								break;
							} else {
								notmyRocksCovered++;// features[4]++;
								break;
							}
						}
					}
				}
				if (CurlingUtils.isPointRock(rockId, event.getAllRocks(), event
						.getRocksInPlay())) {
					if (event.isMyRock(rockId)) {
						points++;// features[0]++;
					} else {
						points--;// features[0]--;
					}
				}
			} else if (CurlingUtils.isInPlay(event.getCurrentPos(rockId))) {
				if (CurlingUtils.isCenterGuard(event.getCurrentPos(rockId))) {
					centerGuard++;// features[6]++;
				} else
					cornerGuard++;// features[5]++;
			}
		}

	}

	public AgentState(int points, int myRocksInHouse, int notmyRocksInHouse,
			int myRocksCovered, int notmyRocksCovered, int cornerGuard,
			int centerGuard) {
		this.points = points;
		this.myRocksInHouse = myRocksInHouse;
		this.notmyRocksInHouse = notmyRocksInHouse;
		this.myRocksCovered = myRocksCovered;
		this.notmyRocksCovered = notmyRocksCovered;
		this.cornerGuard = cornerGuard;
		this.centerGuard = centerGuard;
	}

	public double[] getDoubleArray() {
		double[] stateArray = { points, myRocksInHouse, notmyRocksInHouse,
				myRocksCovered, notmyRocksCovered, cornerGuard, centerGuard };
		return stateArray;
	}

	public String[] getDescription() {
		String[] descriptionArray = { "Constant.", "Match Points.",
				"Num. of Rocks that is mine and in house.",
				"Num. of Rocks that is NOT mine and in house.",
				"Num. of Rocks that is mine and covered.",
				"Num. of Rocks that is NOT mine and covered.",
				"Num. of Cornerguards.", "Num. of Centerguards." };
		return descriptionArray;
	}

	@Override
	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
	}

	@Override
	public String toString() {
		return "Points: \t\t\t\t\t\t" + points
				+ "\nNum. My Rocks in house:\t\t" + myRocksInHouse
				+ "\nNum. Others Rocks in house:\t" + notmyRocksInHouse
				+ "\nNum. My Rocks Covered:\t\t" + myRocksCovered
				+ "\nNum. Others Rocks Covered:\t" + notmyRocksCovered
				+ "\nNum. CornerGuards:\t\t\t" + cornerGuard
				+ "\nNum. CenterGuards:\t\t\t" + centerGuard;
	}

	public void setEstimate(double estimate) {
		this.estimate = estimate;
	}

	public double getEstimate() {
		return estimate;
	}

	protected static  double inHouseScore(double distance) {
		return 1 - distance / IN_HOUSE_CONSTANT;
	}

	public boolean equals(Object that) {
		if (this == that)
			return true;
		if (!(that instanceof AgentState)) {
			return false;
		}
		AgentState thats = (AgentState) that;
		return Arrays.equals(getDoubleArray(), thats.getDoubleArray());
	}

	public int hashCode() {
		int hash = 7;
		double[] ll = getDoubleArray();
		for (double d : ll)
			hash = (int) (31 * hash + d);
		return hash;
	}

}
