package no.skippy.agent.skippy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import no.skippy.agent.event.RoundEndedEvent;
import no.skippy.agent.event.TurnEvent;
import no.skippy.agent.examples.LFunction;
import no.skippy.agent.skippy.api.Action;
import no.skippy.agent.skippy.api.Featurizer;
import no.skippy.agent.skippy.api.State;
import no.skippy.agent.skippy.rlearning.ActionFactory;
import no.skippy.agent.util.CurlingUtils;

/**
 * Curling Agent that uses Linear Target Function when playing.
 * 
 */
public class SkippyAgent extends AbstractAgent {

	protected final int DEFAULT_NUMBER_OF_ACTION = 3;
	
	private int numberOfTopActionToCheck = DEFAULT_NUMBER_OF_ACTION;
	
	
	protected void configure() {
		setAgentName("E1\tSkippyAgent Mod (Experimental Update)");
		useHammer(false);
		//setNormalSuffix("06.05.09_0");// SkippyAgent_15.04.09_0.xml
		setNormalSuffix("15.04.09_0");
		setHammerSuffix("");
		setNumberOfActionsToCheck(DEFAULT_NUMBER_OF_ACTION);
		setWriteOut(false);
	}

	@Override
	public void setup() {
		// Initialized with random weights
		if (features == null) {
			normalFeatures = new AgentStateFeatures(AgentState.LOWER_BOUNDS,
					AgentState.UPPER_BUUNDS, AgentState.NUM_FEATURES);
			hammerFeatures = new AgentStateFeatures(AgentState.LOWER_BOUNDS,
					AgentState.UPPER_BUUNDS, AgentState.NUM_FEATURES);

			// Initialize with the loaded data.
			if (getResource(getAgentData()) != null)
				initialize(normalFeatures, getResource(getAgentData()));

			if (getResource(getAgentDataHammer()) != null)
				initialize(hammerFeatures, getResource(getAgentDataHammer()));
		}

		logNormal.addEntry(logEntry);

		printFeatureWeights(normalFeatures, new AgentState(), false);
		if (isUseHammer()) {
			logHammer.addEntry(logEntry);
			printFeatureWeights(hammerFeatures, new AgentState(), true);
		}
	}

	@Override
	public void onTurn(TurnEvent event) {
		bestAction = new AgentAction(event);
		doBestMove(bestAction, event);
	}

	protected double testActionWithRisk(Action action, TurnEvent event) {
		double actionX = action.getPos().getX();
		int actionSplitTime = action.getSplitTime();

		State testState;
		double riskEstimate = 0;
		for (Action sub : action.getSubActions()) {
			testState = getState(sub, event);
			if (sub.getSplitTime() <= CurlingUtils.TAKEOUT_SPEED) {
				riskEstimate += CurlingUtils.getMissProbability(sub.getPos()
						.getX()
						- actionX, .25)
						* CurlingUtils.getMissProbability(sub.getSplitTime()
								- actionSplitTime, 15)
						* features.getEstimate(testState);
			} else {
				riskEstimate += CurlingUtils.getMissProbability(sub.getPos()
						.getX()
						- actionX, .5)
						* CurlingUtils.getMissProbability(sub.getSplitTime()
								- actionSplitTime, 30)
						* features.getEstimate(testState);
			}
		}
		return riskEstimate;
	}

	/**
	 * 
	 * Sort the sub-actions available from an action by their state estimates.
	 * </p>
	 * If the action does not have any sub-action, the method will return 
	 * a list with only one member, the parentAction.
	 * 
	 * @param parentAction, action to use.
	 * @param event to use for testing the actions..
	 * @return a list of sub action.
	 */
	protected List<Object[]> testTopThreeMoves(Action parentAction, TurnEvent event) {
		
		// since state -> many actions,
		// actions can't be sorted by hashing and sorting state/action
		
		List<Object[]> actionStatePairs = new ArrayList<Object[]>();
		
		State testState;
		for (Action sub : parentAction.getSubActions()) {
			testState = getState(sub, event);
			testState.setEstimate(features.getEstimate(testState));
			actionStatePairs.add(new Object[] {testState, sub});
		}
		
		Collections.sort(actionStatePairs, sorter);

		if (actionStatePairs.isEmpty()) {
			actionStatePairs.add( new Object[] { new AgentState(event), parentAction} );
		}
		
//		ArrayList<Action> topActions = new ArrayList<Action>();
//		if (actionStatePairs.isEmpty()) { // no sub-actions
//			topActions.add(parentAction);  
//		}
//		
//		for (int i = 0; i < actionStatePairs.size()	; i++) {
//			topActions.add((Action) actionStatePairs.get(i)[1]);
//		}
		
//		HashMap<String, Action> actions = new HashMap<String, Action>();
//		ArrayList<State> topState = new ArrayList<State>();

//		bestState = getState(parentAction, event); // current state
//		bestState.setEstimate(features.getEstimate(bestState));
//		
//		State testState;
//		for (Action sub : parentAction.getSubActions()) {
//
//			testState = getState(sub, event);
//			testState.setEstimate(features.getEstimate(testState));
//
//			topState.add(testState);
//			String key = Arrays.toString(testState.getDoubleArray());
//			actions.put(key, sub);
//		}
		
	
//		State[] states = (State[]) topState.toArray(new State[topState.size()]);
//		Arrays.sort(states, stateSorter);

		
		
//		for (State state : states) {
//			topActions
//					.add(actions.get(Arrays.toString(state.getDoubleArray())));
//		}
//		
//		if(topActions.isEmpty()) {
//			topActions.add(parentAction);
//		}
//		
		return actionStatePairs;
	}

	protected void doBestMove(Action bestAction, TurnEvent event) {

		features = getFeaturesFor(event);

		List<Object[]> topActions = testTopThreeMoves(bestAction, event);
		double bestScore = Double.NEGATIVE_INFINITY;

		for (int i = 0; i < topActions.size() && i < numberOfTopActionToCheck; i++) {
			double testActionScore = testActionWithRisk(ActionFactory
					.createRiskAction((AgentAction) topActions.get(i)[1]), event);
			if (testActionScore > bestScore) {
				bestScore = testActionScore;
				bestAction = (Action) topActions.get(i)[1];
				bestState = (State) topActions.get(i)[0];
			}
		}
		
		addStateLog(bestState); // log the state
		bestAction.execute(event); // sets broom properties.
		roundsPlayed++;
	}

	@Override
	public void onLoose(RoundEndedEvent event) {
		trainWeights(-event.getPoints(), 100 * -event.getPoints());
	}

	@Override
	public void onTie(RoundEndedEvent event) {
		if (event.isHasHammer())
			trainWeights(event.getPoints(), 150);
		else
			trainWeights(event.getPoints(), -150);
	}

	@Override
	public void onWin(RoundEndedEvent event) {
		trainWeights(event.getPoints(), 100 * event.getPoints());
	}

	private void initialize(Featurizer features, Object object) {
		LFunction function = (LFunction) object;
		features.setFeatureWeights(function.getWeights());
		((AgentStateFeatures) features).learningRate = function.getConstant();
		roundsPlayed = function.getRoundsPlayed();
		setTrainingCount(features, function.getTrainingCount());
	}

	private Featurizer getFeaturesFor(TurnEvent event) {
		if (event.isHasHammer() && isUseHammer()) {
			hammerRound = true;
			return hammerFeatures;
		} else {
			hammerRound = false;
			return normalFeatures;
		}
	}

	protected void setNumberOfActionsToCheck(int number) {
		numberOfTopActionToCheck = number;
	}
	
	private static StateActionComparator sorter = new StateActionComparator();

	static class StateComparator implements Comparator<State> {

		public int compare(State o1, State o2) {
			State s1 = (State) o1;
			State s2 = (State) o2;

			double diff = s2.getEstimate() - s1.getEstimate();
			if (diff < 0) {
				return -1;
			} else if (diff > 0) {
				return 1;
			} else
				return 0;
		}
	}
	
	static class StateActionComparator implements Comparator<Object[]> {

		public int compare(Object[] o1, Object[] o2) {

			State s1 = (State) o1[0];
			State s2 = (State) o2[0];

			double diff = s2.getEstimate() - s1.getEstimate();
			if (diff < 0) {
				return -1;
			} else if (diff > 0) {
				return 1;
			} else
				return 0;

		}

	}
}
