package no.skippy.agent.skippy.rlearning;

import java.awt.geom.Point2D;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import no.skippy.agent.skippy.AgentAction;
import no.skippy.agent.skippy.api.Action;
import no.skippy.agent.util.CurlingUtils;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;

/**
 * Support class for {@link RLSkippyAgent}. Generates sets of actions and states
 * that is used to kick start the agent.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class ActionFactory {

	protected static String AGENT_ACTION_PROPERTIES = null;
	
	public static final String SELECTED_ACTION = "/no/skippy/agent/skippy/rlearning/selected.properties";
	
	private static final String DEFAULT_AGENT_ACTION_PROPERTIES = "/no/skippy/agent/skippy/rlearning/smallreinforcementActions.properties";
	
	public Map<Integer,Action> actionMap = new HashMap<Integer,Action>();
	
	private static final Log log = JCLoggerFactory
			.getLogger(ActionFactory.class);
	
	static {
		Properties p = new Properties();
		try {
			p.load(ActionFactory.class.getResourceAsStream(SELECTED_ACTION));
			AGENT_ACTION_PROPERTIES = p.getProperty("agentAction");
		} catch (IOException e) {
			log.error("failed to load properties for the factory", e);
			if (AGENT_ACTION_PROPERTIES == null) {
				AGENT_ACTION_PROPERTIES = DEFAULT_AGENT_ACTION_PROPERTIES;
			}
		}
	}

	private static ActionFactory instance = new ActionFactory();

	public ActionFactory() {

		Properties props = new Properties();
		try {
			props.load(ActionFactory.class
					.getResourceAsStream(AGENT_ACTION_PROPERTIES));
		} catch (IOException e) {
			log.error("Failed to load agent props from file: "
					+ AGENT_ACTION_PROPERTIES, e);
			props = null;
		} catch (NullPointerException e) {
			log.error("Failed to find agent props file: "
					+ AGENT_ACTION_PROPERTIES, e);
			props = null;
		}

		int numberOfActions = Integer.parseInt(props.getProperty(
				"numberofActions", "0"));

		for (int actionID = 0; actionID < numberOfActions; actionID++) {
			String name = props.getProperty(String.format("action%s.name", actionID),
					"No description");
			double xpos = Double.parseDouble(props.getProperty(String.format(
					"action%s.xpos", actionID), "0.0"));
			double ypos = Double.parseDouble(props.getProperty(String.format(
					"action%s.ypos", actionID), "0.0"));
			boolean turn = Boolean.parseBoolean(props.getProperty(String
					.format("action%s.outturn", actionID), "0"));
			int split = Integer.parseInt(props.getProperty(String.format(
					"action%s.split", actionID), "3333"));

			int subs = Integer.parseInt(props.getProperty(String.format(
					"action%s.subsCount", actionID), "0"));

			PropertyLoadedAction action = new PropertyLoadedAction(
					new Point2D.Double(xpos, ypos), split, turn, name);
			
			if (subs > 0) {
				for (int sub = 0; sub < subs; sub++) {

					String res = props.getProperty(String.format(
							"action%s.sub%s", actionID, sub),
							"0, 0, 0, 1, 3333, 3333, true, 0, 0");

					String[] r = res.split(",");
					if (r.length == 10) {
						double minX = Double.parseDouble(r[0].trim());
						double minY = Double.parseDouble(r[1].trim());
						double maxX = Double.parseDouble(r[2].trim());
						double maxY = Double.parseDouble(r[3].trim());
						int minSplit = Integer.parseInt(r[4].trim());
						int maxSplit = Integer.parseInt(r[5].trim());
						boolean out = Boolean.parseBoolean(r[6].trim());
						double dx = Double.parseDouble(r[7].trim());
						double dy = Double.parseDouble(r[8].trim());
						double ds = Double.parseDouble(r[9].trim());

						log.debug(String
								.format("Sub: %s %s %s %s %s %s %s %s %s %s ",
										name, minX, minY, maxX, maxY, minSplit,
										maxSplit, out, dx, dy, ds));
						
						// ensure that the values is in right order
						double dminX = Math.min(maxX, minX);
						double dmaxX = Math.max(maxX, minX);
						double dminY = Math.min(maxY, minY);
						double dmaxY = Math.max(maxY, minY);
						
						int dminSplit = Math.min(maxSplit, minSplit);
						int dmaxSplit = Math.max(maxSplit, minSplit);
						
						action.addSubActions(dminX, dminY, dmaxX, dmaxY, dminSplit,
								dmaxSplit, out, dx, dy, ds);
					} else {
						log.warn("Error in propertiy: "
								+ String.format("action%s.subsCount", actionID));
					}
				}
				actionMap.put(actionID, action);
			}
		}
	}

	public static ActionFactory getInstance() {
		return instance;
	}

	/**
	 * Returns a Map of actions.
	 * 
	 * @return set of all available actions.
	 */
	public Map<Integer, Action> getAllActions() {
		return actionMap;
	}

	private static class PropertyLoadedAction extends AgentAction {
		public PropertyLoadedAction(Point2D pos, int splitTime,
				boolean outTurn, String description) {
			super(pos, splitTime, outTurn);
			this.description = description;
		}

		protected void addSubActions(double minX, double minY, double maxX,
				double maxY, int minSplit, int maxSplit, boolean turn,
				double dx, double dy, double ds) {

			for (double x = minX; x < maxX; x += dx) {
				for (double y = minY; y < maxY; y += dy) {
					Point2D pos = new Point2D.Double(x, y);
					for (int split = minSplit; split < maxSplit; split += ds) {
						subActions.add(new PropertyLoadedAction(pos, split,
								turn, description));
					}
				}
			}
		}

		public String getDescription() {
			return description;
		}

		public String toString() {
			return getDescription();
		}
		
		private String description;
	}

	private static class RiskAction extends AgentAction {
		private String description;

		public RiskAction(AgentAction action) {
			super(action.getPos(), action.getSplitTime(), action.isOutTurn());
			this.description = "RiskAction - " + action.getDescription();

			if (action.getSplitTime() <= CurlingUtils.TAKEOUT_SPEED) {
				for (double x = -.25; x <= .25; x += .25) {
					for (int splitT = -15; splitT <= 15; splitT += 15) {
						subActions.add(new AgentAction(new Point2D.Double(
								action.getPos().getX() + x, action.getPos()
										.getY()), action.getSplitTime()
								+ splitT, action.isOutTurn()));
					}
				}
			} else {
				for (double x = -.5; x <= .5; x += .5) {
					for (int splitT = -30; splitT <= 30; splitT += 30) {
						subActions.add(new AgentAction(new Point2D.Double(
								action.getPos().getX() + x, action.getPos()
										.getY()), action.getSplitTime()
								+ splitT, action.isOutTurn()));
					}
				}
			}
		}

		public String getDescription() {
			return description;
		}
	}

	public static RiskAction createRiskAction(AgentAction action) {
		return new RiskAction(action);
	}
}
