package no.skippy.agent.skippy.rlearning;

import java.sql.Statement;
import java.util.Arrays;
import java.util.List;

import no.skippy.agent.skippy.api.Action;
import no.skippy.agent.skippy.api.State;
import no.skippy.agent.skippy.rlearning.db.QDataBase;
import no.skippy.agent.skippy.rlearning.db.RowEntry;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;

public class QLearning {

	private static final Log log = JCLoggerFactory.getLogger(QLearning.class);

	/** SQL handler. */
	private QDataBase qdBase;

	/** Cached row from QDatabase. */
	private RowEntry cachedRow;

	private final double DISCOUNT_FACTOR = 0.9;

	protected final double MAX_LEARNING_RATE = 0.8;

	/** faster lockup when running multiple instances of the QLearning. */
	private boolean lockAccess = false;
	
	private int zeroQvalueUpdated = 0;

	/**
	 * Constructor. Initialize its QDatabase with all actions from
	 * {@link ActionFactory#getAllActions()}
	 */
	public QLearning() {
		qdBase = new QDataBase(ActionFactory.getInstance().getAllActions(),
				lockAccess);
	}

	/**
	 * Setup the QLearning. Connects to database.
	 */
	public void setup() {
		qdBase.connect();
	}

	/**
	 * Exit the QLearning, closes the connection to the database.
	 */
	public void tearDown() {
		qdBase.disconnect();
		zeroQvalueUpdated = 0;
	}

	/**
	 * Select an action using UCB/UCT
	 * 
	 * @param state
	 *            to use.
	 * @return selected action from the state.
	 * @throws NullPointerException
	 *             if state is null.
	 */
	public Action selectAction(State state) {
		if (state == null) {
			throw new NullPointerException("State can't be null");
		}

		List<RowEntry> rows = qdBase.getQSA(state);

		int stateCountUses = 0;
		double averageQSA = 2.0; // avoid 0 average.
		boolean untested = false;

		cachedRow = null;

		double tmpavg = 0.0;
		for (RowEntry row : rows) {
			stateCountUses += row.getCountUsed();
			// DO NOT BREAK OUT (need state used count)
			if (row.getCountUsed() == 0) {
				untested = true;
				cachedRow = row;
			}
			tmpavg += row.getQvalue();
		}

		double maxValue = Double.NEGATIVE_INFINITY;
		double maxQValue = Double.NEGATIVE_INFINITY;
		if (!untested) {

			averageQSA += tmpavg / rows.size();

			double tmpValue;

			for (RowEntry row : rows) {

				tmpValue = row.getQvalue();

				if (maxQValue < tmpValue) {
					maxQValue = tmpValue;
				}

				tmpValue += averageQSA * C(stateCountUses, row.getCountUsed());

				if (tmpValue > maxValue) {
					maxValue = tmpValue;
					cachedRow = row;
				}
			}

		}

		if (cachedRow.getCountUsed() != 0) {
			log
					.info(String
							.format(
									"Q(s,a) value %s, revard %s, average Q(s,a) %s, maxValue %s maxQValue %s",
									cachedRow.getQvalue(), cachedRow
											.getReward(), averageQSA, maxValue,
									maxQValue));
		} else {
			log.info("Select an unteste action.");
		}

		return qdBase.getAction(cachedRow.getActionID());

	}

	/**
	 * Update the Q value using the Q-learning algorithm.
	 * 
	 * @param state
	 *            to update.
	 * @param action
	 *            used on the state.
	 * @param nextState
	 *            state after applying the action.
	 */
	public void updateValue(State state, Action action, State nextState) {
		cache(state, action, cachedRow);

		double oldValue = cachedRow.getQvalue();
		
		double learningRate = getLearningRate(cachedRow.getCountUsed());

		RowEntry next = getMaxRow(nextState);
		double qvalue = oldValue * (1 - learningRate) + learningRate
				* (cachedRow.getReward() + DISCOUNT_FACTOR * next.getQvalue());

		log.info("old value " + oldValue + " NEW VALUE " + qvalue
				+ " using learning rate " + learningRate + " \n");

		// log changes.
		if (oldValue == 0.0 && qvalue != 0.0) {
			zeroQvalueUpdated++;
		}
		
		cachedRow.setQvalue(qvalue);
		cachedRow.setCountUsed(cachedRow.getCountUsed() + 1);
		qdBase.update(cachedRow);
	}

	/**
	 * Set the reward and update the new Q-value.
	 * 
	 * @param state
	 *            to use.
	 * @param action
	 *            to use.
	 * @param reward
	 *            the new reward.
	 */
	public void setReward(State state, Action action, double reward) {
		cache(state, action, cachedRow);

		cachedRow.setReward(reward);

		double oldValue = cachedRow.getQvalue();
		double learningRate = getLearningRate(cachedRow.getCountUsed());

		double qvalue = oldValue * (1 - learningRate) + learningRate
				* (cachedRow.getReward() + DISCOUNT_FACTOR * 0);
		// it is a 'end state, no next state...

		// log changes.
		if (oldValue == 0.0 && qvalue != 0.0) {
			zeroQvalueUpdated++;
		}
		
		log.info("Final State: old value " + oldValue + " NEW VALUE " + qvalue
				+ " using learning rate " + learningRate + " and reward: "
				+ reward);

		cachedRow.setQvalue(qvalue);
		// cachedRow.setCountUsed(cachedRow.getCountUsed() + 1);
		qdBase.update(cachedRow);
	}

	/**
	 * Get the Q(s,a) value.
	 * 
	 * @param state
	 *            to use.
	 * @param action
	 *            to use.
	 * @return Q value of a state-action pair.
	 */
	public double getValue(State state, Action action) {
		cache(state, action, cachedRow);
		return cachedRow.getQvalue();
	}

	/**
	 * Calculate Confidence interval.
	 * <p>
	 * If either timesState or timesAction are less than 1, 0.0 is returned.
	 * </p>
	 * 
	 * @param timesState
	 *            times the state has been played.
	 * @param timesAction
	 *            number of times the action has been tried at the state.
	 * @return the confidence interval.
	 */
	public static double C(double timesState, double timesAction) {
		if (timesState <= 0 || timesAction <= 0)
			return 0;
		return Math.sqrt((2 * Math.log(timesState)) / timesAction);
	}

	/**
	 * Test if the State, Action is cached, if not, get it cached.
	 * 
	 * @param state
	 *            to check.
	 * @param action
	 *            to check.
	 * @param cachedRow
	 *            the cached row.
	 */
	private void cache(State state, Action action, RowEntry cachedRow) {
		if (!cachedRow.belongsTo(Arrays.toString(state.getDoubleArray()),
				qdBase.getActionID(action))
				|| cachedRow == null) {
			cachedRow = qdBase.getQSARow(state, action);
		}
	}

	protected void logProgress(int rounds, int score, int roundswon, String opponent) {
		String update = "Insert into %s values(%s)";
		String values = "%s, %s, %s, %s, %s, %s, '%s', %s";

		values = String.format(values, "null", "now()", rounds, score, qdBase
				.getStateAddedCount(), roundswon, opponent, zeroQvalueUpdated);

		update = String.format(update, "logMatch", values);

		Statement stm = qdBase.getConnection().createStatement();

		qdBase.getConnection().executeUpdate(update, stm);

		log.info(update);
		qdBase.resetStateAddedCount();
	}

	/**
	 * Get the Row ( max_a Q(s,a) ).
	 * 
	 * @param state
	 *            to use.
	 * @return row with highest q-value.
	 */
	private RowEntry getMaxRow(State state) {
		List<RowEntry> rows = qdBase.getQSA(state);
		// TODO: replace the query with this?
		/*
		* select * from qtable where qvalue=(select max(qvalue) 
		*   from qtable where stateid="state");
		*/
		double maxValue = Double.NEGATIVE_INFINITY;
		RowEntry maxRow = null;
		for (RowEntry row : rows) {
			if (row.getQvalue() > maxValue) {
				maxValue = row.getQvalue();
				maxRow = row;
			}
		}
		return maxRow;
	}

	protected double getLearningRate(int countUsed) {
		return (1000.0 * MAX_LEARNING_RATE) / (1000.0 + countUsed);
	}

}
