package no.skippy.agent.skippy.rlearning;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import no.skippy.agent.event.AfterTurnEvent;
import no.skippy.agent.event.RoundEndedEvent;
import no.skippy.agent.event.TurnEvent;
import no.skippy.agent.skippy.AgentState;
import no.skippy.agent.skippy.SkippyAgent;
import no.skippy.agent.skippy.api.Action;
import no.skippy.agent.skippy.api.State;
import no.skippy.agent.util.Storage;

/**
 * Agent based on {@link SkippyAgent}, but uses Reinforcement Learning in an
 * effort to heighten its curling level and shorten 'think'/computation time.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class RLSkippyAgent extends SkippyAgent {

	/** The number of rocks set in a round. (should be reset after each round)- */
	protected int rockSet = 0;

	/** Current action selected */
	protected Action action;

	/** Current state. */
	protected State currentState;

	/**
	 * Expected state after currentState + action. (based on the
	 * {@link bestState}.
	 */
	protected State expectedState;

	protected QLearning qLearning;

	protected List<Integer> pointLog;

	protected boolean logProgress;

	protected boolean useQSAWithHammerState;
	
	private String opponent;

	protected void configure() {
		setAgentName("E3 RLSkippyAgent QLearning (RLSkippy)");
		useHammer(true);
		setNormalSuffix("26.05.09_Trained");
		setHammerSuffix("26.05.09_Trained");
		setNumberOfActionsToCheck(DEFAULT_NUMBER_OF_ACTION);
		setWriteOut(false);

		// configure locally for RLSkippyAgent
		setUseStateWithHammer(true);
		setLogProgress(true);
	}

	@Override
	public void setup() {
		super.setup(); // initialize SkippyAgent LFeatures
		if (qLearning == null) {
			qLearning = new QLearning();
		}
		if (pointLog == null) {
			pointLog = new ArrayList<Integer>();
		}

		pointLog.clear();
		qLearning.setup(); // connects to database.
	}

	public void tearDown(Storage service) {
		super.tearDown(service); // logging could be useful.
		// enable through configure setWriteOut(true)

		if (this.logProgress) {
			int rounds = pointLog.size();
			int score = 0;

			int roundswon = 0;
			for (Integer point : pointLog) {

				if (point > 0) {
					roundswon += 1;
				}

				score += point;
			}
			qLearning.logProgress(rounds, score, roundswon, opponent);
		}
		qLearning.tearDown(); // disconnect database.
	}
	
	@Override
	public void onTurn(TurnEvent event) {
		currentState = getState(event);
		action = qLearning.selectAction(currentState);

		log("Selected action: " + action.getDescription() + " in state: "
				+ Arrays.toString(currentState.getDoubleArray()));
		doBestMove(action, event);
		rockSet++;
		
		opponent = event.getOponent();
	}

	@Override
	public void afterTurn(AfterTurnEvent event) {
		// (estimate next turn state)
		expectedState = getState((AgentState) bestState, Math.min(rockSet, 7),
				event.isHasHammer());
		qLearning.updateValue(currentState, action, expectedState);
	}

	@Override
	public void onLoose(RoundEndedEvent event) {
		qLearning.setReward(currentState, action, -100 * event.getPoints());
		pointLog.add(-event.getPoints());
		super.onLoose(event);
		reset(event);
	}

	@Override
	public void onTie(RoundEndedEvent event) {
		if (event.isHasHammer()) {
			qLearning.setReward(currentState, action, 200);
		} else {
			qLearning.setReward(currentState, action, -200);
		}
		super.onTie(event);
		pointLog.add(event.getPoints());
		reset(event);
	}

	@Override
	public void onWin(RoundEndedEvent event) {
		qLearning.setReward(currentState, action, 100 * event.getPoints());
		super.onWin(event);
		reset(event);
		pointLog.add(event.getPoints());
	}

	private void reset(RoundEndedEvent event) {
		rockSet = 0;
	}

	@Override
	public RLAgentState getState(TurnEvent event) {
		if (useQSAWithHammerState) {
			RLAgentStateHammer state = new RLAgentStateHammer(event, rockSet);
			state.setHammer(event.isHasHammer());
			return state;
		}// else
		return new RLAgentState(event, rockSet);
	}

	@Override
	public RLAgentState getState(Action action, TurnEvent event) {
		if (action != null) {
			action.execute(event);
		}
		if (useQSAWithHammerState) {
			RLAgentStateHammer state = new RLAgentStateHammer(event, rockSet);
			state.setHammer(event.isHasHammer());
			return state;
		}// else
		return new RLAgentState(event, rockSet);
	}

	private RLAgentState getState(AgentState bestState, int rock, boolean hammer) {
		if (useQSAWithHammerState) {
			RLAgentStateHammer state = new RLAgentStateHammer(bestState, rock);
			state.setHammer(hammer);
			return state;
		}// else
		return new RLAgentState(bestState, rock);

	}

	protected void setLogProgress(boolean logProgress) {
		this.logProgress = logProgress;
	}

	protected void setUseStateWithHammer(boolean useQSAWithHammerState) {
		this.useQSAWithHammerState = useQSAWithHammerState;
	}
}
