package no.skippy.agent.skippy.rlearning.db;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import no.skippy.agent.skippy.rlearning.QLearning;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;

/**
 * SQL database wrapper. Keeps a connection to a database and provides access to
 * main SQL services.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class DBConnection {

	/** Log instance. */
	private static Log logger = JCLoggerFactory.getLogger(QLearning.class);

	/** properties for the database, url, user/pass etc. */
	private DBProperties db;

	/** true if the connection to the database is open. */
	private boolean connected = false;

	/** the SQL connection. */
	private Connection conn = null;

	/**
	 * Constructor.
	 * 
	 * @param properties
	 *            to use for database connection.
	 */
	public DBConnection(DBProperties properties) {
		this.db = properties;
	}

	/**
	 * Attempts to establish a connection to the given database given by
	 * {@link DBProperties}.
	 * 
	 * @throws SQLException
	 *             - if a database access error occurs.
	 */
	public boolean connect() {
		try {
			Class.forName("com.mysql.jdbc.Driver").newInstance();
			conn = DriverManager
					.getConnection(db.url, db.userName, db.password);
			connected = true;
			logger.info("Connected to Database.(" + db.url + ")"
					+ " using the tables: " + db.actionTable + " and "
					+ db.qtable);
		} catch (Exception e) {
			logger.error("Connect error: ", e);
			connected = false;
		}
		return connected;
	}

	/**
	 * Releases this DBConnection object's database and JDBC resources
	 * immediately instead of waiting for them to be automatically released.
	 * <p>
	 * Calling the method close on a DBConnection object that is already closed
	 * is a no-op.
	 * </p>
	 * 
	 * @throws SQLException
	 *             - if a database access error occurs.
	 */
	public boolean close() {
		try {
			conn.close();
			connected = false;
			logger.info("Disconneted from Database");
		} catch (SQLException e) {
			logger.error("Close error: ", e);
		}
		return connected;
	}

	/**
	 * Get connection status.
	 * 
	 * @return true if a connection to the given database is established.
	 */
	public boolean isConnected() {
		return connected;
	}

	/**
	 * Executes the given SQL statement, which returns a single ResultSet
	 * object.
	 * 
	 * @param sql
	 *            an SQL statement to be sent to the database, typically a
	 *            static SQL SELECT statement.
	 * @param statement
	 *            used for executing a static SQL statement and returning the
	 *            results it produces.
	 * @return a ResultSet object that contains the data produced by the given
	 *         query; never null
	 * 
	 * @throws SQLException
	 *             - if a database access error occurs or the given SQL
	 *             statement produces anything other than a single ResultSet
	 *             object
	 */
	public ResultSet executeQuery(String sql, Statement statement) {

		ResultSet set = null;
		if (!connected) {
			logger.error(String.format(ERR_CONNECTED, "execute query"));
		} else {
			try {
				set = statement.executeQuery(sql);
			} catch (SQLException e) {
				logger.error("Failed to execute query", e);
			}
		}
		return set;
	}

	/**
	 * Executes the given SQL statement, which may be an <code>INSERT</code>,
	 * <code>UPDATE</code>, or <code>DELETE</code> statement or an SQL statement
	 * that returns nothing, such as an SQL DDL statement.
	 * 
	 * @param sql
	 *            an SQL <code>INSERT</code>, <code>UPDATE</code> or
	 *            <code>DELETE</code> statement or an SQL statement that returns
	 *            nothing.
	 * @param statement
	 *            used for executing a static SQL statement and returning the
	 *            results it produces.
	 * @return either the row count for <code>INSERT</code>, <code>UPDATE</code>
	 *         or <code>DELETE</code> statements, or <code>0</code> for SQL
	 *         statements that return nothing.
	 * @throws SQLException
	 *             - if a database access error occurs or the given SQL
	 *             statement produces a ResultSet object
	 */
	public int executeUpdate(String sql, Statement statement) {
		int returnValue = -1;
		if (!connected) {
			logger.error(String.format(ERR_CONNECTED, "execute query"));
		} else {
			try {
				returnValue = statement.executeUpdate(sql);
			} catch (SQLException e) {
				logger.error("Faild to execute update", e);
			}
		}
		return returnValue;
	}

	/**
	 * Creates a Statement object for sending SQL statements to the database.
	 * 
	 * @return a new default Statement object
	 * @throws SQLException
	 *             - if a database access error occurs
	 */
	public Statement createStatement() {
		Statement statement = null;
		if (!connected) {
			logger.error(String.format(ERR_CONNECTED, "create Statement"));
		} else {
			try {
				statement = conn.createStatement();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return statement;
	}

	/**
	 * Lock write access.
	 * 
	 * @param statement
	 *            used for executing a <code>LOCK TABLES</code>
	 * @param table
	 *            to lock.
	 */
	public void lockTable(Statement statement, String table) {
		try {
			statement.execute(String.format("LOCK TABLES %s WRITE", table));
		} catch (SQLException e) {
			logger.error("Faild to LOCK TABLES", e);
		}
	}

	/**
	 * Unlock a statements locked tables.
	 * 
	 * @param statement
	 *            used for executing a <code>UNLOCK TABLES</code>.
	 */
	public void unlock(Statement statement) {
		try {
			statement.execute("UNLOCK TABLES");
		} catch (SQLException e) {
			logger.error("Failed to UNLOCK table", e);
		}
	}

	/**
	 * Select Count(*) from <code>sql</code>.
	 * 
	 * @param sql
	 *            SQL to use.
	 * @param statement
	 *            to use.
	 * @return row count.
	 */
	public int getRowCount(String sql, Statement statement) {
		int count = 0;

		sql = String.format("Select Count(*) from %s", sql);

		ResultSet rs = executeQuery(sql, statement);
		try {
			rs.first();
			count = rs.getInt(1);
		} catch (SQLException e) {
			e.printStackTrace();
		}
		return count;
	}

	/**
	 * Convert received result set to a {@link RowEntry} objects.
	 * 
	 * @param set
	 *            to convert.
	 * @return a list of {@link RowEntry}
	 * @throws SQLException
	 *             - if a database access error occurs.
	 */
	public List<RowEntry> getRows(ResultSet set) {
		List<RowEntry> entries = new ArrayList<RowEntry>();
		try {
			while (set.next()) {
				RowEntry entry = new RowEntry();
				entry.stateID = set.getString("stateID");
				entry.actionID = set.getInt("actionID");
				entry.qvalue = set.getDouble("qValue");
				entry.reward = set.getDouble("reward");
				entry.countUsed = set.getInt("countUsed");
				entries.add(entry);
			}
		} catch (SQLException e) {
			logger.error(ERR_ROWENTRY, e);
		}
		return entries;
	}

	/**
	 * Test for an empty ResultSet
	 * 
	 * @param resultSet
	 *            to test.
	 * @return <code>true</code> if the ResultSet is empty.
	 */
	public boolean isEmpty(ResultSet resultSet) {
		boolean found = false;
		try {
			found = resultSet.next();
			resultSet.beforeFirst();
			return !found;

		} catch (SQLException e) {
			logger.error("Faild to check result set.", e);
		}
		throw new IllegalStateException(ERR_RESULTSET);
	}

	private String ERR_CONNECTED = "Tried to %s, but was not connected. Check your mysql.properties";
	private String ERR_RESULTSET = "Could not check the ResultSet";
	private String ERR_ROWENTRY = "Faild to create RowEntry of resultset";
}
