package no.skippy.agent.skippy.rlearning.db;

import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import no.skippy.agent.skippy.api.Action;
import no.skippy.agent.skippy.api.State;
import no.skippy.agent.skippy.rlearning.QLearning;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;

/**
 * Class for handling the SQL operation needed by the {@link QLearning} class.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class QDataBase {

	/** log instance. */
	private static Log logger = JCLoggerFactory.getLogger(QLearning.class);

	/** SQL database connection. */
	private DBConnection db;

	/** Properties */
	private DBProperties dbProp;

	/** Map actionID to action objects. */
	private Map<Integer, Action> actionMap = new HashMap<Integer, Action>();

	/** Map actionDescription to actionID (SQL table key). */
	private Map<String, Integer> aKeyMap = new HashMap<String, Integer>();
	
	private int stateAddedCount = 0;
	
	/** True if update/access should be restricted. */
	
	private boolean lockAccess = true;

	/**
	 * Constructor.
	 * 
	 * @param allActions
	 *            to use.
	 */
	public QDataBase(Map<Integer, Action> allActions) {
		for (Integer key : allActions.keySet()) {
			Action action = allActions.get(key);
			actionMap.put(key, action);
			aKeyMap.put(action.getDescription(), key);
		}
		dbProp = new DBProperties();
		db = new DBConnection(dbProp);
	}
	
	/**
	 * @param allActions to use.
	 * @param lockAccess <code>True</code> if access should be restricted. (prevents adding duplicates).
	 */
	public QDataBase(Map<Integer, Action> allActions, boolean lockAccess ) {
		this(allActions);
		this.lockAccess = lockAccess;
	}

	/**
	 * Create database connection.
	 */
	public void connect() {
		db.connect();
		if (db.isConnected()) {
			checkActions();
		}
	}

	/**
	 * Close database connection.s
	 */
	public void disconnect() {
		db.close();
	}

	/**
	 * Get action using actionID from SQL table.
	 * 
	 * @param key
	 *            actionID
	 * @return action.
	 */
	public Action getAction(int key) {
		return actionMap.get(key);
	}

	/**
	 * Get actionID from an action
	 * 
	 * @param action
	 *            object.
	 * @return actionID from SQL table.
	 */
	public int getActionID(Action action) {
		return aKeyMap.get(action.getDescription());
	}

	/**
	 * Get the table row for state and action.
	 * 
	 * @param state
	 *            to use.
	 * @param action
	 *            to use.
	 * @return row with all properties where state/action is equal to the
	 *         parameters.
	 */
	public RowEntry getQSARow(State state, Action action) {

		String stateID = Arrays.toString(state.getDoubleArray());
		int actionID = getActionID(action);

		Statement statement = db.createStatement();
		String sql = String.format(SELECT_ALL_FROM + ROWS_FROM_TABLE_WHERE
				+ AND, dbProp.qtable, dbProp.stateKey, stateID,
				dbProp.actionKey, actionID);
		ResultSet set = db.executeQuery(sql, statement);
		
		if (db.isEmpty(set)) { 
			addStateActions(stateID, statement);
			set = db.executeQuery(sql, statement);
		}

		if (set == null) {
			logger.error(ERRNULL);
		} else {
			return db.getRows(set).get(0);
		}
		throw new IllegalStateException(String.format(ERR1, stateID, actionID));
	}

	/**
	 * Get all action using a state.
	 * @param state to use.
	 * @return all possible actions available from a state.s
	 */
	public List<RowEntry> getQSA(State state) {

		List<RowEntry> entries = new ArrayList<RowEntry>();
		String stateID = Arrays.toString(state.getDoubleArray());

		Statement statement = db.createStatement();
		String sql = String.format(SELECT_ALL_FROM + ROWS_FROM_TABLE_WHERE,
				dbProp.qtable, dbProp.stateKey, stateID);

		ResultSet set = db.executeQuery(sql, statement);

		if (db.isEmpty(set)) {
			addStateActions(stateID, statement);
			set = db.executeQuery(sql, statement);
		}

		if (set == null) {
			logger.error(ERRNULL);
		} else {
			entries = db.getRows(set);
		}

		return entries;
	}

	/**
	 * Update a row in the QTable.
	 * 
	 * @param row
	 *            to update (with updated values).
	 */
	public void update(RowEntry row) {
		String sql = String.format(UPDATE_TABLE_SET_VALUES_WHERE + AND,
				dbProp.qtable, row.getSet(), dbProp.stateKey, row.stateID,
				dbProp.actionKey, row.actionID);

		Statement statement = db.createStatement();
		db.executeUpdate(sql, statement);
	}

	/**
	 * Check that action exist in the action table.
	 */
	private void checkActions() {
		Statement statement = db.createStatement();
		
		if (lockAccess) {
			db.lockTable(statement, dbProp.actionTable);
		}

		int count = db.getRowCount(String.format(ALL_FROM, dbProp.actionTable),
				statement);

		if (count == 0) {
			for (Integer actionID : actionMap.keySet()) {
				String sql = String.format(INSERT_INTO_TABLE_VALUES,
						dbProp.actionTable, String.format(ACTION_VALUES,
								actionID, actionMap.get(actionID)));
				db.executeUpdate(sql, statement);
			}
		}
		if (lockAccess) {
			db.unlock(statement);
		}
	}

	/**
	 * Add State,actions tuples.
	 */
	private void addStateActions(String stateID, Statement statement) {
		if (lockAccess) {
			db.lockTable(statement, dbProp.qtable);
		}
		
		int count = db.getRowCount(String.format(ROWS_FROM_TABLE_WHERE,
				dbProp.qtable, dbProp.stateKey, stateID), statement);

		if (count == 0) {
			logger.info("Adding new action entries for state: " + stateID);
			createStateActionEntries(stateID, statement);
		}

		if (lockAccess) {
			db.unlock(statement);
		}
	}
	
	private void createStateActionEntries(String stateID, Statement statement) {

		for (int actionID : actionMap.keySet()) {
			RowEntry row = new RowEntry(stateID, actionID);

			String sql = String.format(INSERT_INTO_TABLE_VALUES, dbProp.qtable,
					String.format(QTABLE_VALUES, row.values()));
			db.executeUpdate(sql, statement);
			stateAddedCount++;
		}
	}

	/**
	 * @return count of the new entries in QTable.
	 */
	public int getStateAddedCount() {
		return stateAddedCount;
	}
	
	/**
	 *	Reset the new entry count. 
	 */
	public void resetStateAddedCount() {
		stateAddedCount = 0;
	}
	
	public DBConnection getConnection() {
		return db;
	}
	
	private static final String SELECT_ALL_FROM = "SELECT * FROM";
	private static final String ROWS_FROM_TABLE_WHERE = " %s WHERE %s='%s'";
	private static final String AND = " AND %s='%s'";
	private static final String ALL_FROM = " %s";
	private static final String INSERT_INTO_TABLE_VALUES = "Insert into %s values(%s)";
	private static final String UPDATE_TABLE_SET_VALUES_WHERE = "UPDATE %s SET %s WHERE %s='%s'";
	private static final String ACTION_VALUES = "%s, '%s'";
	private static final String QTABLE_VALUES = "'%s',%s,%s,%s,%s";
	private static final String ERR1 = "Failed to retrieve data using stateID: %s and actionID: %s. Check that the actionID "
			+ "match ID's in Action table.";
	private static final String ERRNULL = "A null error that should never have happend.";

}
