package no.skippy.agent.util;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import no.skippy.agent.BasicAgent;
import no.skippy.sim.control.ReadOnlyRockCollection;
import no.skippy.sim.model.RockComparable;

/**
 * Utility class for curling. Provides a collection of useful methods that can
 * be used by {@link BasicAgent} implementations.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class CurlingUtils {

	/** Distance from Back-line to Hog. 21+6 feet (converted to meter) */
	public static final float BACK_2_HOG = feetToMeter(21 + 6);

	/** Distance from Tee to Back-line. 6 feet (converted to meter) */
	public static final float BACK_2_TEE = feetToMeter(6);

	/** Distance from Tee to Hack. 6+6+21+72+21 feet (converted to meter) */
	public static final float FAR_HACK_2_TEE = feetToMeter(6 + 6 + 21 + 72 + 21);

	/**
	 * Distance from Tee to Hog (on player's end). 21+72 feet (converted to
	 * meter)
	 */
	public static final float FAR_HOG_2_TEE = feetToMeter(21 + 72);

	/** Distance from Tee to Hog (converted to meter). */
	public static final float NEAR_HOG_2_TEE = feetToMeter(21);

	/** Distance from Hack to Back-line. 6 feet (converted to meter) */
	public static final float HACK_2_BACK = feetToMeter(6);

	/** Distance from Hack to Hog. 6+6+21 feet (converted to meter) */
	public static final float HACK_2_HOG = feetToMeter(6 + 6 + 21);

	/**
	 * Distance from far Hog-line to near Hog-line. 72 feet (converted to meter)
	 */
	public static final float HOG_2_HOG = feetToMeter(72);

	/** Distance from Tee to Hog (near the house). 21 feet (converted to meter) */
	public static final float HOG_2_TEE = feetToMeter(21);

	/** Distance from Center-line to edge. 6+1 feet (converted to meter) */
	public static final float SIDE_2_CENTER = feetToMeter(7);

	/** House radius (converted to meter) */
	public static final float HOUSE_RAD = feetToMeter(6);

	/** 12 feet radius (converted to meter) */
	public static final float HOUSE_12 = feetToMeter(6);

	/** 8 feet radius (converted to meter) */
	public static final float HOUSE_8 = feetToMeter(4);

	/** 4 feet radius (converted to meter) */
	public static final float HOUSE_4 = feetToMeter(2);
	
	public static final int TAKEOUT_SPEED = 3150;

	public static final float ROCK_RAD;

	/** Tee position. **/
	private static final Point2D TEE;

	static {
		ROCK_RAD = feetToMeter(0.5);
		TEE = new Point2D.Double(0, 0);
	}

	/**
	 * private constructor.
	 */
	private CurlingUtils() {}

	/**
	 * In house.
	 * 
	 * @param pos
	 *            of a rock.
	 * @return true if the rock is in house.
	 */
	public final static boolean isInHouse(Point2D pos) {
		return pos.distance(TEE) < HOUSE_RAD + ROCK_RAD;
	}

	/**
	 * Inside 12 foot.
	 * 
	 * @param pos
	 *            of a rock.
	 * @return true if the rock is inside 12 feet circle. (red)
	 */
	public final static boolean isIn12(Point2D pos) {
		return pos.distance(TEE) < HOUSE_12 - ROCK_RAD;
	}

	/**
	 * Inside 8 foot circle.
	 * 
	 * @param pos
	 *            of a rock.
	 * @return true if the rock is inside the 8 foot circle. (white)
	 */
	public final static boolean isIn8(Point2D pos) {
		return pos.distance(TEE) < HOUSE_8 - ROCK_RAD;
	}

	/**
	 * Inside the 4 foot circle.
	 * 
	 * @param pos
	 *            of a rock.
	 * @return true if the rock is inside the 4 foot circle. (blue)
	 */
	public final static boolean isIn4(Point2D pos) {
		return pos.distance(TEE) < HOUSE_4 - ROCK_RAD;
	}

	/**
	 * Rock is playable.
	 * 
	 * @param pos
	 *            of a rock.
	 * @return true if the rock is between hog line and back line and side
	 *         lines.
	 */
	public final static boolean isInPlay(Point2D pos) {
		return -pos.getY() < BACK_2_TEE + ROCK_RAD // do not cross back line
				&& pos.getY() < NEAR_HOG_2_TEE - ROCK_RAD // has crossed hog line
				&& pos.getX() < SIDE_2_CENTER - ROCK_RAD // not out right
				&& -pos.getX() < SIDE_2_CENTER - ROCK_RAD; // not out left
	}

	/**
	 * Free Guard Zone.
	 * 
	 * @param pos
	 *            of a rock.
	 * @return true if the rock is in free guard zone.
	 */
	public final static boolean isInFGZ(Point2D pos) {
		return -pos.getY() < -ROCK_RAD // do not cross Tee line
				&& pos.getY() < NEAR_HOG_2_TEE - ROCK_RAD // has crossed hog line
				&& pos.getX() < SIDE_2_CENTER - ROCK_RAD // not out right
				&& -pos.getX() < SIDE_2_CENTER - ROCK_RAD // not out left
				&& pos.distance(TEE) > HOUSE_RAD + ROCK_RAD; // not in house
	}

	/**
	 * A corner guard is a rock in front of the house that is not in the
	 * four-foot zone.
	 * 
	 * @param pos
	 *            of a rock.
	 * @return true if the rock is a corner guard.
	 */
	public final static boolean isCornerGuard(Point2D pos) {
		boolean isCornerguard = isInPlay(pos);
		isCornerguard &= !isInHouse(pos);
		isCornerguard &=
			pos.getX() < -feetToMeter(4) ||
			pos.getX() > feetToMeter(4);
		
		return isCornerguard;
	}
	
	/**
	 * A guard that is placed directly on the centre line, in front of the house.  
	 * @param pos of a rock.
	 * @return true if the rock is a center guard.
	 */
	public final static boolean isCenterGuard(Point2D pos) {
		boolean isCenterGuard = isInPlay(pos);
		isCenterGuard &= !isInHouse(pos);
		isCenterGuard &=
			pos.getX() > -feetToMeter(4) &&
			pos.getX() < feetToMeter(4);
		
		return isCenterGuard;
	}
	
	/**
	 * Check if a rock take a point.
	 * 
	 * @param rockid16
	 *            to check.
	 * @param rocks
	 *            collection.
	 * @param rocksInPlay
	 *            check, if null all rocks will be checked.
	 * @return true if the rock (rockid16) is a point rock.
	 */
	public final static boolean isPointRock(int rockid16,
			ReadOnlyRockCollection rocks, Set<Integer> rocksInPlay) {
		Point2D rockPos = rocks.getCurrentPos(rockid16);
		if (!isInHouse(rockPos)) {
			return false; // no reason to test further.
		}

		if (rocksInPlay == null) {
			rocksInPlay = new HashSet<Integer>();
			for (int i = 0; i < 16; i++)
				rocksInPlay.add(i);
		}

		boolean takePoint = true;
		boolean isDark = isDark(rockid16);
		double rockDist = rockPos.distance(0, 0);
		for (int rid : rocksInPlay) {
//			int rid = rocksInPlay.get(i); // rock candidate

			if (rid == rockid16) { // same rock, ignore
				continue;
			} else if (!takePoint) { // not a point rock, escape.
				break;
			}

			if (isDark && !isDark(rid)) {// light rid,
				takePoint &= rockDist < rocks.getCurrentPos(rid).distance(TEE);
			} else if (!isDark && isDark(rid)) { // dark rid
				takePoint &= rockDist < rocks.getCurrentPos(rid).distance(TEE);
			}
		}
		return takePoint;
	}

	/**
	 * Get the current points ratio between dark and light player. One of the
	 * player will always have zero.
	 * 
	 * @param rocks
	 *            positions.
	 * @param rocksInPlay
	 *            rocks to check.
	 * @return array with the points of dark and light player. Dark is index 0.
	 */
	public final static int[] getPoints(ReadOnlyRockCollection rocks,
			Set<Integer> rocksInPlay) {
		int score = 0;
		int[] points = new int[2];
		List<RockComparable> sortedRocks = new ArrayList<RockComparable>();

		for (Integer i16 : rocksInPlay) {
			RockComparable rock = new RockComparable(i16, rocks
					.getCurrentPos(i16));
			sortedRocks.add(rock);
		}

		Collections.sort((List<RockComparable>) sortedRocks);

		if (!sortedRocks.isEmpty()) {
		if (sortedRocks.get(0).isDark()) {
			int i = 0;
			while (i < sortedRocks.size() - 1 &&
					sortedRocks.get(i).isDark() && sortedRocks.get(i).inHouse()) {
				score++;
				i++;
			}
			points[0] = (score);
			points[1] = (0);
		} else {
			int i = 0;
			while (i < sortedRocks.size() -1 && 
					!(sortedRocks.get(i).isDark())&& sortedRocks.get(i).inHouse()) {
				score++;
				i++;
			}
			points[0] = (0);
			points[1] = (score);
		}}
		return points;
	}

	/**
	 * Determine the colour of a rock.
	 * 
	 * @param i16
	 *            index for a rock.
	 * @return true if the rock is dark.
	 */
	public final static boolean isDark(int i16) {
		return i16 % 2 == 0;
	}

	/**
	 * Convert feet to meter.
	 * 
	 * @param feet
	 *            to convert.
	 * @return meter.
	 */
	public final static float feetToMeter(final double feet) {
		return (float) (feet * 0.3048);
	}

	/**
	 * Convert meter to feet.
	 * 
	 * @param meter
	 *            to convert.
	 * @return feet.
	 */
	public final static float meeterTofoot(float meter) {
		return (float) (meter / 0.3048);
	}
	
	/**
	 * Get the probability of missing a shot by x meters, using sd standard deviation
	 * 
	 * @param x
	 *            meters to miss by.
	 * @param sd
	 *            standard deviation to use in calculation.
	 * @return probability of miss.
	 */
	public final static double getMissProbability(double x, double sd){
		return Math.exp(-(x*x/2*sd*sd))/(sd*Math.sqrt(2*Math.PI));
	}
	
	/**
	 * Convert rock color and index per color to index per set.
	 * 
	 * @param isDark
	 * @param idx8
	 *            [0-7]
	 * @return [0-15]
	 */
	public static int toIdx16(final boolean isDark, final int idx8) {
		return 2 * idx8 + (isDark ? 0 : 1);
	}
	
	public static final int toIdx8(final int idx16) {
		return idx16 / 2;
	}

}
