package no.skippy.sim.control;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;

import javax.swing.ActionMap;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileFilter;

import no.skippy.sim.util.GuiCreator;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;
import org.jcurl.core.ui.FileNameExtensionFilter;
import org.jdesktop.application.Action;
import org.jdesktop.application.Application;
import org.jdesktop.application.ApplicationContext;
import org.jdesktop.application.ResourceMap;
import org.jdesktop.application.Task.BlockingScope;

/**
 * Collection of actions with relations to the main Application. e.g. load resources (file), reset,
 * save, export, import.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class ApplicationActions {
	
	private static Log log = JCLoggerFactory.getLogger(ApplicationActions.class);
	
	/** about panel. */
	private JDialog about = null;
	
	/** all actions in this object. */
	private ActionMap actionMap;
	
	private ApplicationContext appContext;
	private ResourceMap resourceMap = null;
	private SimulatorApp mainApp;
	private FileNameExtensionFilter jcxzPat;
	
	public ApplicationActions(SimulatorApp simulatorApp) {
		this.mainApp   = simulatorApp;
		appContext     = Application.getInstance().getContext();
		actionMap      = appContext.getActionMap(this);
		resourceMap    = appContext.getResourceMap(ApplicationActions.class);
		jcxzPat        = createFileFilter("fileFilterJcxz", "jcz", "jcx");
	}

	@Action
	public void fileOpen() {
		final JFileChooser chooser = createJcxChooser(new File(""), "openFileChooser");
		
		if (chooser.showOpenDialog(mainApp.getMainFrame()) == JFileChooser.APPROVE_OPTION) {
			final File file = chooser.getSelectedFile();
			try {
				mainApp.setDocument(file.toURI().toURL());
			}catch (final MalformedURLException e) {
//				crazy
				log.error("File.toURI().toURL() failed", e);
			}catch (final IOException e) {
				log.error("can't open \"" +  file + "\"", e);
			}
		}
	}
	
	@Action(block = BlockingScope.COMPONENT)
	public void showAboutBox() {
		if (about == null) {
			about = createAboutBox(mainApp.getMainFrame());
		}
		mainApp.show(about);
	}
	
	@Action
	public void closeAboutBox() {
		if (about != null) {
			about.setVisible(false);
			about = null;
		}
	}
	
	public JMenu getAppMenu() {
		final String[] viewMenuActionNames = { "showAboutBox", "---", "quit"};
//		more to come later?
		JMenu menu = GuiCreator.createMenu("skippy", viewMenuActionNames, actionMap);
		resourceMap.injectComponents(menu);
		return menu;
	}
	
	public JMenu getFileMenu() {
		final String[] viewMenuActionNames = { "fileOpen"};
		JMenu menu = GuiCreator.createMenu("File", viewMenuActionNames, actionMap);
		resourceMap.injectComponents(menu);
		return menu;
	}
	
	/**
	 * Create a simple about box JDialog that displays the standard
	 * Application resources, like {@code Application.title} and
	 * {@code Application.description}. The about box's labels and fields
	 * are configured by resources that are injected when the about box is
	 * shown (see SingleFrameApplication#show). The resources are defined in
	 * the application resource file: resources/DocumentExample.properties.
	 * 
	 * From:
	 * https://appframework.dev.java.net/downloads/AppFramework-1.03-src.zip
	 * DocumentExample
	 * @param frame 
	 */
	private JDialog createAboutBox(JFrame frame) {
		 JPanel panel = new JPanel(new GridBagLayout());
	        panel.setBorder(new EmptyBorder(0, 28, 16, 28)); // top, left, bottom, right
	        JLabel titleLabel = new JLabel();
	        titleLabel.setName("aboutTitleLabel");
	        GridBagConstraints c = new GridBagConstraints();
	        initGridBagConstraints(c);
	        c.anchor = GridBagConstraints.WEST;
	        c.gridwidth = GridBagConstraints.REMAINDER;
	        c.fill = GridBagConstraints.HORIZONTAL;
	        c.ipady = 32;
	        c.weightx = 1.0;
	        panel.add(titleLabel, c);
	        String[] fields = {"description", "version", "vendor", "home"};
	        for (String field : fields) {
	            JLabel label = new JLabel();
	            label.setName(field + "Label");
	            JTextField textField = new JTextField();
	            textField.setName(field + "TextField");
	            textField.setEditable(false);
	            textField.setBorder(null);
	            initGridBagConstraints(c);
	            //c.anchor = GridBagConstraints.BASELINE_TRAILING; 1.6 ONLY
	            c.anchor = GridBagConstraints.EAST;
	            panel.add(label, c);
	            initGridBagConstraints(c);
	            c.weightx = 1.0;
	            c.gridwidth = GridBagConstraints.REMAINDER;
	            c.fill = GridBagConstraints.HORIZONTAL;
	            panel.add(textField, c);
	        }
	        JButton closeAboutButton = new JButton();
	        closeAboutButton.setAction( getAction("closeAboutBox"));
	        initGridBagConstraints(c);
	        c.anchor = GridBagConstraints.EAST;
	        c.gridx = 1;
	        panel.add(closeAboutButton, c);
	       
	        JDialog dialog = new JDialog(frame);
	        dialog.setName("aboutDialog");
	        dialog.add(panel, BorderLayout.CENTER);
	        dialog.setBackground(Color.white);
	        dialog.setResizable(false);
	        resourceMap.injectComponents(dialog);
	        return dialog;
	}
	
	private JFileChooser createJcxChooser(final File base, final String name) {
		return createFileChooser(base, name, jcxzPat);
	};
	
	private JFileChooser createFileChooser(final File base,
			final String resourceName, final FileFilter filter) {
		final JFileChooser fc = new JFileChooser(base);
		fc.setName(resourceName);
		fc.setMultiSelectionEnabled(false);
		fc.setAcceptAllFileFilterUsed(true);
		fc.setFileFilter(filter);
		resourceMap.injectComponents(fc);
		return fc;
	}
	
	private FileNameExtensionFilter createFileFilter(
			final String resourceName, final String... extensions) {
		
		final String key = resourceName + ".description";
		final String desc = resourceMap.getString(key);
		return new FileNameExtensionFilter(desc == null ? key : desc,
				extensions);
	}
	
//	helpers
	private void initGridBagConstraints(final GridBagConstraints c) {
		c.anchor = GridBagConstraints.CENTER;
		c.fill = GridBagConstraints.NONE;
		c.gridwidth = 1;
		c.gridheight = 1;
		c.gridx = GridBagConstraints.RELATIVE;
		c.gridy = GridBagConstraints.RELATIVE;
		c.insets = zeroInsets;
		c.ipadx = 4; // not the usual default
		c.ipady = 4; // not the usual default
		c.weightx = 0.0;
		c.weighty = 0.0;
	};
	
	 /* A convenience method that looks up the javax.swing.Action 
     * object with the specified name in the Application ActionMap.
     */
    private javax.swing.Action getAction(String actionName) {
        return actionMap.get(actionName);
    }
    
    private static final Insets zeroInsets = new Insets(0, 0, 0, 0);

}
