/*
 * CurlAI, aka Skippy AI http://org.ntnu.no/curlai/ Copyright (C) 2008-2009
 * CurlAI
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

package no.skippy.sim.control;

import java.awt.geom.Point2D;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.BoundedRangeModel;
import javax.swing.SwingUtilities;

import no.skippy.agent.util.CurlingUtils;
import no.skippy.sim.model.RockCollection;

import org.apache.commons.logging.Log;
import org.jcurl.core.api.Rock;
import org.jcurl.core.api.RockSet;
import org.jcurl.core.api.Unit;
import org.jcurl.core.api.RockType.Pos;
import org.jcurl.core.api.RockType.Vel;
import org.jcurl.core.log.JCLoggerFactory;
import org.jcurl.core.ui.BroomPromptModel;
import org.jcurl.core.ui.ChangeManager;
import org.jcurl.core.ui.BroomPromptModel.HandleMemento;
import org.jcurl.core.ui.BroomPromptModel.SplitMemento;
import org.jcurl.core.ui.BroomPromptModel.XYMemento;
import org.jdesktop.application.Action;
import org.jdesktop.application.Task;

/** * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a> */
public class RockBroomControl implements BroomControls, RockSender {

	private static Log log = JCLoggerFactory.getLogger(RockBroomControl.class);

	public static final int ROUND_END = -1;
	
	private BroomPromptModel broomModel;
	private RockCollection rCollection;
	private ChangeManager changer;
	private boolean waitingOnAgent = false;
	private SimulatorApp app;

	private static final int SET_BROOM_PROP_WAIT = 25;

	private int sleepTime = SET_BROOM_PROP_WAIT;

	private PropertyChangeSupport pcs = new PropertyChangeSupport(this);

	public RockBroomControl(BroomPromptModel broomModel,
			RockCollection rockCollection, SimulatorApp app) {
		super();
		this.broomModel = broomModel;
		this.rCollection = rockCollection;
		this.app = app;
	}

	// implementing BroomControls interface.

	public Point2D getBroom() {
		Point2D pp = broomModel.getBroom();
		return new Point2D.Double(pp.getX(), pp.getY());
	}

	public boolean getOutTurn() {
		return broomModel.getOutTurn();
	}

	public BoundedRangeModel getSplitTimeMillies() {
		return broomModel.getSplitTimeMillis();
	}

	public void setBroom(final Point2D pos) {
		final XYMemento pre = new XYMemento(broomModel, broomModel.getBroom());
		final XYMemento post = new XYMemento(broomModel, pos);

		invokeLater(new Runnable() {
			public void run() {
				getChanger().undoable(pre, post);
			}
		});
		sleep(sleepTime);
	}

	public void setOutTurn(final boolean outTurn) {
		final HandleMemento pre = new HandleMemento(broomModel, broomModel
				.getOutTurn());
		final HandleMemento post = new HandleMemento(broomModel, outTurn);

		invokeLater(new Runnable() {
			public void run() {
				getChanger().undoable(pre, post);
			}
		});
		sleep(sleepTime);
	}

	public void setSplitTimeMillies(int splitTime) {
		final SplitMemento pre = new SplitMemento(broomModel, broomModel
				.getSplitTimeMillis().getValue());
		final SplitMemento post = new SplitMemento(broomModel, splitTime);
		invokeLater(new Runnable() {
			public void run() {
				getChanger().undoable(pre, post);
			}
		});
		sleep(sleepTime);
	}

	// implementing RockSender interface

	public boolean isWaiting() {
		return waitingOnAgent;
	}

	public void reset(final int idx16) {
		if (isWaiting()) {
			log.warn("Reset the rocks while waiting on an agent.");
		}

		invokeAndWait(new Runnable() {
			public void run() {
				Task<Void, Void> v = app.getResetTask(); // hm, could it be an easier way
				// to reset without requiring I/O?
				v.execute();
				
				while (!v.isDone())
					sleep(5);
				broomModel.setIdx16(idx16);
				waitingOnAgent = false;
			}

		});

		rocksPlayed.clear();
	}

	public void setRock(int current, int next, int rockCount) {
		setRock(current, next, rockCount, false);
	}

	public void setRock(int current, int next, int rockCount,
			boolean interactive) {
		if (interactive) {
			this.current = current;
			this.next = next;
			this.rockCount = rockCount;
			setWaiting(interactive);
		} else {
			set(current, next, rockCount);
		}
	}

	private int current, next, rockCount;

	private void set(final int current, final int next, int rockCount) {
		rocksPlayed.put(current + "", current);
		rocksInPlay.add(current);

		final RockSet<Pos> cPos = rCollection.getCurves().getCurrentPos(); // current
		final RockSet<Pos> iPos = rCollection.getCurves().getInitialPos(); // initial
		final RockSet<Vel> cVel = rCollection.getCurves().getCurrentVel(); // current
		final RockSet<Vel> iVel = rCollection.getCurves().getInitialVel(); // initial

		// backup broom's rock position.
		final double x = cPos.getRock(current).getX();
		final double y = cPos.getRock(current).getY();
		final double a = cPos.getRock(current).getA();

		boolean sentAFGZrockOut = false;
		tmpRocksInPlay = new HashSet<Integer>();

		if (current >=0) {
			rocksInPlay.add(current);
		}
		for (Integer i16 : rocksInPlay) {
			if (CurlingUtils.isInPlay(cPos.getRock(i16).p())) {
				tmpRocksInPlay.add(i16);
			} else if (rockCount < 5) {
				if (current == i16) {
					continue;
				}
				if (CurlingUtils.isInFGZ(iPos.getRock(i16).p())) {
					sentAFGZrockOut = true;
					tmpRocksInPlay.add(i16);
					log.info("Player sendt rock(s) that was FGZ out, "
							+ "the rock will be moved back. ");
				}
			}
		}

		// remove rock that was not valid (5 rock FGZ protect rule).
		if (sentAFGZrockOut && tmpRocksInPlay.contains(current)) {
			tmpRocksInPlay.remove(current);
//			tmpRocksInPlay.remove(tmpRocksInPlay.indexOf(current));
		}

		rocksInPlay = tmpRocksInPlay; // update playable rocks.

		// preserve the rocks current positions.
		// not doing this will reset recursive collision.
		for (int i = RockSet.ROCKS_PER_SET - 1; i >= 0; i--) {
			Rock<Pos> cu = cPos.getRock(i);
			xp[i] = cu.getX();
			yp[i] = cu.getY();
			ar[i] = cu.getA();
		}

		final boolean rockMoveOut = sentAFGZrockOut;
		
		// update all rock positions.
		for (int i = RockSet.ROCKS_PER_SET - 1; i >= 0; i--) {
			final int i16 = i;
			invokeLater(new Runnable() {
				public void run() {
					if (!rockMoveOut) { // current rock was valid.
						if (rocksPlayed.containsKey(i16 + "")) {
							if (!CurlingUtils.isInPlay(new Point2D.Double(
									xp[i16], yp[i16]))) {
								rocksOut.add(i16);
								// log.info("Rock " + i16 + ": is out of game");
								Point2D out = moveOut(i16);
								iPos.getRock(i16).setLocation(out.getX(),
										out.getY(), 0);
							} else {
								iPos.getRock(i16).setLocation(xp[i16], yp[i16],
										ar[i16]);
							}// TODO bad bad bad coding.
						} else {
							iPos.getRock(i16).setLocation(xp[i16], yp[i16],
									ar[i16]);
						}
					} 
//						else {
//						// do nothing, the rocks positions should not be
//						// altered. (current rock was not valid, and is
//						// therefore removed from game.
//					}
					cVel.getRock(i16).setLocation(0, 0, 0);
					iVel.getRock(i16).setLocation(0, 0, 0);
				}
			});
		}

		// swap broom rock, and set its position.
		invokeAndWait(new Runnable() {
			public void run() {
				if (next > -1) {
					broomModel.setIdx16(next);
				} else {
					broomModel.setIdx16(current);
				}

				if (rockMoveOut) { // move the rock out of play zone.
					// probably a nicer way to do this?
					Point2D out = moveOut(current);
					cPos.getRock(current)
							.setLocation(out.getX(), out.getY(), 0);
					iPos.getRock(current)
							.setLocation(out.getX(), out.getY(), 0);
				} else {

					if (!CurlingUtils.isInPlay(new Point2D.Double(x, y))) {
						Point2D out = moveOut(current);
						iPos.getRock(current).setLocation(out.getX(),
								out.getY(), 0);
						cPos.getRock(current).setLocation(out.getX(),
								out.getY(), 0);
					} else {
						iPos.getRock(current).setLocation(x, y, a);
						cPos.getRock(current).setLocation(x, y, a);
					}
				}
			}
		});

		rocksOut.clear();

		// reset broom
		if (next != ROUND_END) { // except at last rock.
			setSplitTimeMillies(broomSplitReset);
			setBroom(broomPosReset);
			setOutTurn(broomOutTurnReset);
		}
	}
	
	

	/**
	 * @param isWaiting
	 *            set wait for human user input.
	 */
	public void setWaiting(boolean isWaiting) {
		boolean oldValue = isWaiting();
		pcs.firePropertyChange("waiting", oldValue,
				this.waitingOnAgent = isWaiting);
	}

	/**
	 * Set a rock from GUI.
	 */
	@Action(enabledProperty = "waiting")
	public void interactiveSendRock() {
		
		log.info("Agent, set the rock: " + current + " > " + next);
		new Thread(new Runnable() {
			public void run() {
				
				humanBroom = (Point2D) getBroom().clone();
				log.info("Human broom properties: " + humanBroom);
				humanSplit = getSplitTimeMillies().getValue();
				humanTurn = getOutTurn();
				
				set(current, next, rockCount);
				setWaiting(false);
			}
		}).start();
	}

	/**
	 * Needed by the appFramework for supporting enable/disable
	 * {@link RockBroomControl#interactiveSendRock()} action by setting
	 * {@link RockBroomControl#setWaiting(boolean)}.
	 * 
	 * @param listener
	 *            to use.
	 */
	public void addPropertyChangeListener(PropertyChangeListener listener) {
		pcs.addPropertyChangeListener(listener);
	}

	/**
	 * Needed by the appFramework for supporting enable/disable
	 * {@link RockBroomControl#interactiveSendRock()} action by setting
	 * {@link RockBroomControl#setWaiting(boolean)}.
	 * 
	 * @param listener
	 *            to use.
	 */
	public void removePropertyChangeListener(PropertyChangeListener listener) {
		pcs.removePropertyChangeListener(listener);
	}

	protected void setSleepAmount(int sleep) {
		this.sleepTime = sleep;
	}

	/**
	 * Get a list of rock in play.
	 * 
	 * @return a list of all playable rocks.
	 */
	protected Set<Integer> getRocksInPlay() {
		return rocksInPlay;
	}

	private void invokeAndWait(Runnable task) {
		try {
			SwingUtilities.invokeAndWait(task);
		} catch (InterruptedException e) {
			log.error("invokeAndWait (interrupt): ", e);
		} catch (InvocationTargetException e) {
			log.error("invokeAndWait (invocation): ", e);
		}
	}

	private void invokeLater(Runnable task) {
		SwingUtilities.invokeLater(task);
	}

	private void sleep(int millis) {
		try {
			Thread.sleep(millis);
		} catch (InterruptedException e) {
			log.error("Sleeping: ", e);
		}
	}

	private ChangeManager getChanger() {
		return ChangeManager.getTrivial(changer);
	}

	private boolean isDark(int i16) {
		return i16 % 2 == 0;
	}

	@SuppressWarnings("unused")
	private int toIdx8(final int idx16) {
		return idx16 / 2;
	}

	private Point2D moveOut(int idx16) {
		return new Point2D.Double(
				(isDark(idx16) ? -1 : 1) * Unit.f2m(7.25 + 4), Unit
						.f2m(31 - 1.2 * idx16));
	}

	private Set<Integer> rocksInPlay = new HashSet<Integer>();
	private final Map<String, Integer> rocksPlayed = new HashMap<String, Integer>();
	private final HashSet<Integer> rocksOut = new HashSet<Integer>();
	private Set<Integer> tmpRocksInPlay = null;
	private final double[] xp = new double[RockSet.ROCKS_PER_SET];
	private final double[] yp = new double[RockSet.ROCKS_PER_SET];
	private final double[] ar = new double[RockSet.ROCKS_PER_SET];
	private final Point2D broomPosReset = new Point2D.Double(0.0, 0.0);
	private final int broomSplitReset = 3700;
	private final boolean broomOutTurnReset = true;

	protected Point2D getHumanBroom() {
		return humanBroom;
	}

	protected int getHumanSplit() {
		return humanSplit;
	}

	protected boolean getHumanTurn() {
		return humanTurn;
	}
	
	private Point2D humanBroom;
	private int humanSplit;
	private boolean humanTurn;

}
