package no.skippy.sim.control;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.io.IOException;
import java.net.URL;
import java.util.Arrays;
import java.util.Set;

import javax.swing.ActionMap;
import javax.swing.DefaultListModel;
import javax.swing.JComponent;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.ListModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import no.skippy.agent.Agent;
import no.skippy.agent.examples.BullsEyeAgent;
import no.skippy.agent.examples.ExampleAgent;
import no.skippy.agent.examples.ExampleHumanAgent;
import no.skippy.agent.examples.LinearTargetAgent;
import no.skippy.agent.examples.ScottAgent;
import no.skippy.agent.skippy.SkippyAgent;
import no.skippy.agent.skippy.SkippyAgentBasic;
import no.skippy.agent.skippy.SkippyAgentBasicLocked;
import no.skippy.agent.skippy.SkippyAgentHammer;
import no.skippy.agent.skippy.SkippyAgentV2;
import no.skippy.agent.skippy.SkippyAgentV3;
import no.skippy.agent.skippy.SkippyAka;
import no.skippy.agent.skippy.rlearning.RLSkippyAgent;
import no.skippy.sim.model.RockCollection;
import no.skippy.sim.util.LoadJCurlResource;
import no.skippy.sim.view.NewSimulatorDialog;
import no.skippy.sim.view.ScorePanel;
import no.skippy.sim.view.SimulatorBirdView;
import no.skippy.sim.view.SimulatorView;
import no.skippy.sim.view.StatusBar;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;
import org.jcurl.core.ui.TrajectoryBroomPromptWrapper;
import org.jdesktop.application.Action;
import org.jdesktop.application.Resource;
import org.jdesktop.application.ResourceMap;
import org.jdesktop.application.SingleFrameApplication;
import org.jdesktop.application.Task;
import org.jdesktop.application.TaskMonitor;

/**
 * Main application controller.
 * 
 * @see <a href="https://appframework.dev.java.net/intro/index.html">Swing
 *      Application Framework</a>.
 */

public class SimulatorApp extends SingleFrameApplication {

	private static final Log log = JCLoggerFactory
			.getLogger(SimulatorApp.class);

	@Resource
	private int windowHeight;

	@Resource
	private int windowWidth;

	@Resource
	private double currentTime;

	private URL document = null;

	private static URL initialScene = null;

	/** Changes in the simulator. */
	private final ChangeListenerManager cl = new ChangeListenerManager(this);

	private ApplicationActions applicationActions = null;

	private SimulatorViewController viewController = null;

	private RockBroomControl rockBroomControl = null;

	/** Broom control/model wrapper. */
	private final TrajectoryBroomPromptWrapper broomModel;

	/** Zooming on the {@link SimulatorView}. */
	private ZoomActions zoomActions = null;

	/** Simulator speed slider. */
	private JSlider simulatorDelaySlider = null;
	private final int SIMULATOR_MIN_DELAY = 5;
	private final int SIMULATOR_MAX_DELAY = 100;
	private final int SIMULATOR_INIT_DELAY = 25;

	/** Main curling view. */
	private SimulatorView simView = null;

	/** Birdview of the curling view. */
	private SimulatorBirdView birdSimView = null;

	/** View listing the round score. */
	@Resource
	private ScorePanel scorePanel = null;

	/** All rocks, their positions, etc. */
	private final RockCollection rocks;

	/** Model with the round scores. */
	private ListModel scoreListModel;

	/** Not already running a simulator. */
	private boolean simulatorReady = true;

	/** Distorter that makes the simulator nondeterministic. */
	private BroomDistoter distorter = new SimpleDistorter();

	/** Task 'Factory, load/save class instances to disk. */
	private StorageService storageService = new StorageService(this);

	/** Agent available in NewSimulatorDialog list. */
	private Agent[] allAgentsTypes = new Agent[] { 
		    new SkippyAgentBasic(),
			new SkippyAgent(),
			new SkippyAgentV2(), 
			new RLSkippyAgent(),
			new ExampleHumanAgent(), 
			new ScottAgent(), 
			new ExampleAgent(),
	// new StorageAgent(),
	};

	public SimulatorApp() {
		broomModel = new TrajectoryBroomPromptWrapper();
		rocks = new RockCollection();
		scoreListModel = new DefaultListModel();
	}

	@Override
	protected void startup() {
		// Initialise views.
		simView = new SimulatorView(rocks, broomModel);
		birdSimView = new SimulatorBirdView(simView);
		scorePanel = new ScorePanel(scoreListModel);
		//		

		zoomActions = new ZoomActions(simView); // enable zooming from view.

		simulatorDelaySlider = new JSlider(JSlider.HORIZONTAL,
				SIMULATOR_MIN_DELAY, SIMULATOR_MAX_DELAY, SIMULATOR_INIT_DELAY);
		simulatorDelaySlider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				rockBroomControl
						.setSleepAmount(simulatorDelaySlider.getValue());
			}
		});

		// Initialise controllers
		applicationActions = new ApplicationActions(this);
		viewController = new SimulatorViewController(simView, rocks);
		rockBroomControl = new RockBroomControl(broomModel, rocks, this);

		new MouseRockHandler(rocks);

		getMainFrame().setJMenuBar(createMenu()); // create file menus

		final JComponent content = new JPanel();
		content.setLayout(new BorderLayout());

		simView.setPreferredSize(new Dimension(windowHeight, windowWidth));

		final JComponent controlPanel = new JPanel(new BorderLayout());
		birdSimView.setPreferredSize(new Dimension(windowHeight / 2,
				windowWidth / 2));

		ActionMap aMap = getContext().getActionMap(rockBroomControl);

		controlPanel.setBackground(Color.white);
		controlPanel.add(scorePanel, BorderLayout.NORTH);
		controlPanel.add(birdSimView, BorderLayout.CENTER);

		StatusBar statusBar = new StatusBar(new TaskMonitor(getContext()), aMap
				.get("interactiveSendRock"), simulatorDelaySlider);

		simView.setBackground(Color.white);
		content.add(statusBar, BorderLayout.SOUTH);
		content.add(controlPanel, BorderLayout.EAST);
		content.add(simView, BorderLayout.CENTER);

		ResourceMap resource = getContext().getResourceMap();
		resource.injectFields(this);

		show(content);
		zoomActions.view12Foot(); // close up zoom.
	}

	@Action(enabledProperty = "simulatorReady")
	public Task<Void, Void> newSimulator() {
		setSimulatorReady(false);

		NewSimulatorDialog dialog = new NewSimulatorDialog(getContext(),
				getMainFrame(), allAgentsTypes);

		Agent agent1 = null, agent2 = null;
		int rounds, repeat;
		boolean exitOnComplete;
		if (dialog.showDialog(getMainFrame()).isRunSimulatorApproved()) {
			try {
				agent1 = (Agent) dialog.getAgent1().newInstance();
				agent2 = (Agent) dialog.getAgent2().newInstance();
			} catch (InstantiationException e) {
				log.error("Error, when creating new agent instance", e);
			} catch (IllegalAccessException e) {
				log.error("IllegalAccessException, when creating new agent", e);
				e.printStackTrace();
			}
			rounds = dialog.getNumberOfRounds();
			repeat = dialog.getNumberOfMatch();
			exitOnComplete = dialog.isExitOnCompleteSelected();
			dialog.dispose();
			dialog = null;

		} else {
			setSimulatorReady(true);
			return null;
		}

		if (agent1 != null && agent2 != null) {
			return createSimulatorTask(agent1, agent2, rounds, repeat,
					exitOnComplete);
		}
		return null;
	}

	@Override
	protected void ready() {
		final URL tmp = document;
		document = null;
		try {
			setDocument(tmp, true);
		} catch (final IOException e) {
			log.warn("Couldn't load '" + tmp + "'.", e);
		}

		if (initAgents) {
			setSimulatorReady(false);
			RunSimulatorAndExit task = createSimulatorTask(initAgent1,
					initAgent2, initRounds, initMatchs, initExitOnComplete);
			getContext().getTaskService().execute(task);
		}
	};

	/**
	 * Argument initialisation.
	 * <p>
	 * Valid argument is of:<br/>
	 * args=
	 * "Agent1Name|Agent1Class, Agent1Name|Agent1Class, numberOfRounds, numberOfMatch"
	 * </p>
	 * Running from maven example:<br/>
	 * <code>mvn exec:java -Dexec.mainClass="no.skippy.App" -Dexec.args="SkippyAgent, RLSkippyAgent, 10, 1"</code>
	 */
	protected void initialize(final String[] args) {
		if (args.length > 0) {
			initializeWithArgs(args);
		}

		final Class<?> mc = this.getClass();
		final ResourceMap r = getContext().getResourceMap();
		initialScene = mc.getResource("/" + r.getResourcesDir()
				+ r.getString("Application.defaultDocument"));
		document = initialScene;
	}

	private void setDocument(final URL document, boolean load)
			throws IOException {

		final URL old = this.document;
		this.firePropertyChange("document", old, this.document = document);

		if (load) {
			getContext().getTaskService().execute(
					new LoadJCurlResource(this, document, currentTime));
		}
	}

	protected void setDocument(final URL document) throws IOException {
		this.setDocument(document, true);
	}

	/**
	 * @return rocks change listener.
	 */
	public ChangeListenerManager getChangeListenerManager() {
		return cl;
	}

	/**
	 * @return simulator view.
	 */
	public SimulatorView getSimView() {
		return simView;
	}

	/**
	 * @return view controller {@link SimulatorViewController}.
	 */
	public SimulatorViewController getViewController() {
		return viewController;
	}

	public RockCollection getRockCollection() {
		return rocks;
	}

	protected RockSender getRockSender() {
		return rockBroomControl;
	}

	public BroomControls getAgentControls() {
		return rockBroomControl;
	}

	public void setModified(boolean modified) {
		log.debug("setModified not handled");
	}

	protected RockBroomControl getRockBroomControl() {
		return rockBroomControl;
	}

	/**
	 * Get rocks in play.
	 * 
	 * @return list containing all rocks in play.
	 */
	protected Set<Integer> getRocksInPlay() {
		return rockBroomControl.getRocksInPlay();
	}

	protected BroomDistoter getDistorter() {
		return distorter;
	}

	protected void setDistorter(BroomDistoter distorter) {
		this.distorter = distorter;
	}

	public DefaultListModel getScoreListModel() {
		return (DefaultListModel) scoreListModel;
	}

	public int getSimulatorDelay() {
		return simulatorDelaySlider.getValue();
	}

	public StorageService getStorageService() {
		return storageService;
	}

	private JMenuBar createMenu() {
		JMenuBar menuBar = new JMenuBar();
		menuBar.add(applicationActions.getAppMenu());
		menuBar.add(applicationActions.getFileMenu());
		menuBar.add(zoomActions.getMenu());

		JMenu menu = new JMenu();
		menu.setText("Simulator");
		menu.add(getContext().getActionMap().get("newSimulator"));
		menuBar.add(menu);
		return menuBar;
	}

	/**
	 * newSimulator enabler.
	 * 
	 * @return true if no simulation is running.
	 */
	public boolean isSimulatorReady() {
		return simulatorReady;
	}

	/**
	 * Enable/Disable newSimulator/ Start new Simulator.
	 * 
	 * @param simulatorReady
	 *            to set.
	 */
	public void setSimulatorReady(boolean simulatorReady) {
		boolean oldValue = isSimulatorReady();
		firePropertyChange("simulatorReady", oldValue,
				this.simulatorReady = simulatorReady);
	}

	protected Task<Void, Void> getResetTask() {
		return new LoadJCurlResource(this, document, currentTime);
	}

	/**
	 * Change the application frame title to:
	 * <p/>
	 * Application.title :::: agent1Name vs agent2Name <br>
	 * if either argument is null, the frame title is set to default
	 * Application.title.
	 * 
	 * @param agent1
	 *            opponent 1.
	 * @param agent2
	 *            opponent 2.
	 */
	private void setFrameTitle(Agent agent1, Agent agent2) {
		String frameTitle = getContext().getResourceMap().getString(
				"Application.title");

		if (agent1 != null && agent2 != null)
			frameTitle += " :::: %s vs %s";
		if (getMainFrame() != null)
			getMainFrame().setTitle(
					String.format(frameTitle, agent1.getAgentName(), agent2
							.getAgentName()));
	}

	private RunSimulatorAndExit createSimulatorTask(Agent agent1, Agent agent2,
			int rounds, int matchs, boolean exitOnComplete) {
		String name1 = agent1.getAgentName();
		String name2 = agent2.getAgentName();
		log.info(String.format(WELCOME, name1, name2, rounds, matchs));
		agent1.loadResources(getStorageService());
		agent2.loadResources(getStorageService());
		getStorageService().loadResource();
		return new RunSimulatorAndExit(this, agent1, agent2, rounds, matchs,
				exitOnComplete);
	}

	/**
	 * Valid arguments:
	 * <ul>
	 * <li><code>-list</code> <br/>
	 * or</li>
	 * <li><code>-agents agent1 agent2</code><br/>
	 * or with extra arguments</li>
	 * <li><code>-agents agent1 agent2 [-k -r# -m #]</code> where:</li>
	 * <ul>
	 * <li><b>-k</b> Keep Application running after Simulator is done.</li>
	 * <li><b>-r</b> <i>number</i> of rounds to run.</li>
	 * <li><b>-m</b> <i>number</i> of matchs to run.</li>
	 * </ul>
	 * </ul>
	 * 
	 * @param args
	 *            to use
	 */
	private void initializeWithArgs(String[] args) {
		int i = 0;
		String arg;

		while (i < args.length && args[i].startsWith("-")) {
			arg = args[i++];
			if (arg.equalsIgnoreCase("-list")) {
				printAllAgents();
				System.exit(0);
			} else if (arg.equalsIgnoreCase("-agents")) {
				if (i + 1 < args.length) {
					String agent1 = args[i++];
					String agent2 = args[i++];
					log.info(agent1 + " " + agent2);
					parseAgents(agent1, agent2, args);
				}
			} else if (arg.equalsIgnoreCase("-r")) {
				if (i < args.length)
					initRounds = parseNumber(args, i++);
			} else if (arg.equalsIgnoreCase("-m")) {
				if (i < args.length)
					initMatchs = parseNumber(args, i++);
			} else if (arg.equalsIgnoreCase("-k")) {
				initExitOnComplete = false;
			}
		}

		if (initAgents) {
			String agent1Name = initAgent1.getAgentName();
			String agent2Name = initAgent2.getAgentName();
			log.info(String.format(SUCCESS, argsToString(args)));
			log.info(String.format(LOAD_AGENT, agent1Name, agent2Name));
			log.info(String.format(ROUND_MATCH, initRounds, initMatchs));
		} else {
			printUsage();
		}
	}

	private void parseAgents(String agent1, String agent2, String[] args) {
		try {
			for (Agent agent : allAgentsTypes) {
				if (agent1.equalsIgnoreCase(agent.getAgentName())) {
					initAgent1 = agent.getClass().newInstance();
				}
				if (agent2.equalsIgnoreCase(agent.getAgentName())) {
					initAgent2 = agent.getClass().newInstance();
				}
			}
		} catch (InstantiationException e) {} catch (IllegalAccessException e) {}

		// else try class loading.
		if (initAgent1 == null) {
			initAgent1 = loadAgentByClassName(agent1);
		}

		if (initAgent2 == null) {
			initAgent2 = loadAgentByClassName(agent2);
		}

		if (initAgent1 != null && initAgent2 != null) {
			initAgents = true;
			return;
		} else {
			log.error(String.format(ERROR1, argsToString(args)));
		}
	}

	private Agent loadAgentByClassName(String agent) {
		throw new NotYetImplementedException(
				"Can't load agent using classpath+className");
	}

	private int parseNumber(String[] args, int pos) {
		int number = 0;
		try {
			number = Integer.parseInt(args[pos]);
		} catch (NumberFormatException e) {
			log.error(String.format(PARSE_ERROR, args[pos - 1]), e);
			return 1;
		}
		return number;
	}

	private void printUsage() {
		log.info("usage: java App [-list | -agents agent1 agent2] "
				+ "[-r # -m # -k#]");
	}

	private void printAllAgents() {
		log.info("Available SkippySim Agents:");
		for (int i = 0; i < allAgentsTypes.length; i++) {
			Agent a = allAgentsTypes[i];
			log.info(String.format("%s: %s (%s)", i, a.getAgentName(), a
					.isInterActive() ? "Human" : "Computer"));
		}
	}

	private String argsToString(String[] args) {
		return Arrays.toString(args).replaceAll("[,|\\[|\\]]", "");
	}

	private Agent initAgent1 = null;
	private Agent initAgent2 = null;
	private boolean initAgents = false;
	private int initRounds = 10;
	private int initMatchs = 1;
	private boolean initExitOnComplete = true;

	private static String WELCOME = "Starting new simulator, %s vs. %s. Playing %s rounds a %s time(s).";
	private static String LOAD_AGENT = "Initialized with agents: %s and %s.";
	private static String ROUND_MATCH = "Playing %s rounds and %s match.";
	private static String SUCCESS = "Simulator initialized with arguments:  %s";
	private static String ERROR1 = "java App %s is not a valid App command. see java App -help";
	private static String PARSE_ERROR = "Failed to parse %s ";

	/**
	 * Capsule the SimulatorTask.<p/>
	 * Adds the feature <i>'exit on complete'</i> to the Simulator Task. <p/>
	 * Sets the Frame title to the agents playing.
	 */
	class RunSimulatorAndExit extends SimulatorTask {

		private boolean exit;

		public RunSimulatorAndExit(SimulatorApp app, Agent agent1,
				Agent agent2, int numberOfRounds, int match, boolean exit) {
			super(app, agent1, agent2, numberOfRounds, match);
			this.exit = exit;
			setFrameTitle(agent1, agent2);
		}

		protected void succeeded(Void ignored) {
			super.succeeded(null);
			if (exit)
				exit();
		}

	}
}
