package no.skippy.sim.control;

import java.awt.Toolkit;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import no.skippy.agent.Agent;
import no.skippy.agent.event.AfterTurnEvent;
import no.skippy.agent.event.RoundEndedEvent;
import no.skippy.agent.event.TurnEvent;
import no.skippy.agent.event.RoundEndedEvent.Result;
import no.skippy.sim.model.RockComparable;
import no.skippy.sim.model.SimulatorScore;

import org.apache.commons.logging.Log;
import org.jcurl.core.api.RockSet;
import org.jcurl.core.api.RockType.Pos;
import org.jcurl.core.log.JCLoggerFactory;
import org.jdesktop.application.Task;

public class SimulatorTask extends Task<Void, Void> {

	/** log instance. */
	private static final Log log = JCLoggerFactory
			.getLogger(SimulatorTask.class);

	/** out formats. */
	private final static String TOTAL = "Totals: %s (red): %s - %s (yellow): %s";
	/** out formats. */
	private final static String LOG_MATCH = "Match Statistics after %s match played:\t%s %s - %s %s (%s - %s).";
	/** out formats. */
	private final static String ROUND_START = "Round %s starts.\n";
	/** out formats. */
	private final static String ROUND_COMPLETED = "Round %s completed.";
	/** out formats. */
	private final static String MATCH_NUMBER = "Match %s: ";
	/** out formats. */
	private final static String MATCH_TIE = "%s\tTie.";
	/** out formats. */
	private final static String MATCH_ENDED = "%s \tWinner was: %s with %s. (%s)";
	/** out formats. */
	private final static String MATCH_STATUS = "Match %s (of %s).   Round %s.";
	/** out formats. */
	private final static String RUN_EXCEPTION = "Somthing fishy happend. Contact your local curling support line.";

	/** Time to sleep while waiting on interactive (human) agent. */
	private final static int INTERACTIVE_SLEEP = 200;

	/** Amount to wait after an agent has set a rock. */
	private final static int AGENT_SLEEP = 100;

	/** Simulator Application. */
	private SimulatorApp simulator;

	/** Number of rounds to play. */
	private int numberOfRounds;

	/** Repeat 'number of rounds <b>a</b> times. */
	private int numberOfMatch;

	/** Current and initial rock positions. */
	private ReadOnlyRockCollection rockCollection;

	/** Dark (red) agent. */
	private Agent agent1;

	/** Light (yellow) agent. */
	private Agent agent2;

	/** Current Agent playing- */
	private Agent selectedAgent = null;

	/** List of round score. */
	private List<Integer> agent1Score = new ArrayList<Integer>();

	/** List of round score. */
	private List<Integer> agent2Score = new ArrayList<Integer>();

	/** Times the agent has won. */
	private int agent1MatchWinnerCount = 0;
	
	/** Total points an agent got over n match. */
	private int agent1MatchPointsTotal = 0;

	/** Times the agent has won. */
	private int agent2MatchWinnerCount = 0;
	
	private int agent2MatchPointsTotal = 0;

	/**
	 * Creates a new curling match. (Number of match (repeats) is 1)
	 * 
	 * @param app
	 *            .
	 * @param agent1
	 *            player. (dark/red)
	 * @param agent2
	 *            player. (light/yellow)
	 * @param numberOfRounds
	 *            number of rounds to play.
	 */
	public SimulatorTask(SimulatorApp app, Agent agent1, Agent agent2,
			final int numberOfRounds) {
		super(app);
		this.simulator = app;
		if (agent1 == null || agent2 == null) {
			throw new UnsupportedOperationException("Not all players set");
		}

		this.agent1 = agent1; // always red? (dark)
		this.agent2 = agent2;
		this.numberOfRounds = numberOfRounds;
		rockCollection = new ReadOnlyRockCollection(app.getRockCollection());

		this.numberOfMatch = 1;
	}

	/**
	 * Creates a new curling match.
	 * 
	 * @param app
	 *            .
	 * @param agent1
	 *            player. (dark/red)
	 * @param agent2
	 *            player. (light/yellow)
	 * @param numberOfRounds
	 *            number of rounds to play.
	 * @param numberOfRepeats
	 *            number of match (repeats) to play.
	 */
	public SimulatorTask(SimulatorApp app, Agent agent1, Agent agent2,
			final int numberOfRounds, final int numberOfRepeats) {
		this(app, agent1, agent2, numberOfRounds);

		if (numberOfRepeats < 1) {
			log.warn("Repeats set to 1. The number of repeats called was < 1 "
					+ numberOfRepeats);
		}

		this.numberOfMatch = Math.max(1, numberOfRepeats);
	}

	private void setup() {

		simulator.getScoreListModel().clear();
		agent1.setup();
		agent2.setup();

		if (Math.random() < 0.5) {
			this.selectedAgent = agent1;
			setRockOrder(darkRockOrder);
		} else {
			this.selectedAgent = agent2;
			setRockOrder(lightRockOrder);
		}
	}

	private void teardown() {
		StorageService service = simulator.getStorageService();
		agent1.tearDown(service);
		agent2.tearDown(service);
		service.saveResources();

		if (agent1Score.size() != agent2Score.size()) {
			log.error("Agent score list not of equal size");
		}
		
		log.info("Results:");
		for (int i = 0; i < agent1Score.size(); i++) {
			printRoundScore(i, agent1Score.get(i), agent2Score.get(i));
		}

		log.info(format(TOTAL, agent1.getAgentName(), sum(agent1Score), agent2
				.getAgentName(), sum(agent2Score)));

		agent1Score.clear();
		agent2Score.clear();

		log.info(format(LOG_MATCH,
				(agent1MatchWinnerCount + agent2MatchWinnerCount), 
				agent1.getAgentName(), 
				agent1MatchWinnerCount,agent2MatchWinnerCount, 
				agent2.getAgentName(),
				agent1MatchPointsTotal, agent2MatchPointsTotal));
	}

	@Override
	protected Void doInBackground() {

		try {
			int progress = 0;

			int match = 1;
			while (match <= numberOfMatch) {

				setMessage(format(MATCH_NUMBER, match));
				log.info(format(MATCH_NUMBER, match));

				// setup the agents.
				setup();
				
				int roundNr = 0;
				int extraRounds = 0;
				while (roundNr < numberOfRounds + extraRounds) {

					setMessage(format(MATCH_STATUS, match, numberOfMatch, roundNr + 1));
					log.info(format(ROUND_START, (roundNr + 1)));

					// reset round
					int[] rockOrder = getRockOrder();
					int totalProgress = (numberOfRounds + extraRounds)
							* rockOrder.length * numberOfMatch;
					simulator.getRockSender().reset(rockOrder[0]);

					// play a round
					for (int rockID = 0; rockID < rockOrder.length - 1; rockID++) {
						progress++;

						// change rock on broom.
						simulator.getRocksInPlay().add(rockOrder[rockID]);

						// set rock
						playRock(selectedAgent, rockOrder[rockID],
								rockOrder[rockID + 1], rockID, roundNr,
								numberOfRounds);

						// switch to next agent
						selectedAgent = nextAfter(selectedAgent);
						
						if (rockOrder[rockID + 1] == ROUND_END) {
							break;
						}
					}

					computeScore();

					// add round score to score list
					simulator.getScoreListModel().addElement(
							new SimulatorScore((roundNr + 1), agent1Score
									.get(roundNr), agent2Score.get(roundNr)));

					// send win/loose/tie event
					notifyPlayers(rockCollection, roundNr);

					int diff = agent1Score.get(roundNr)
							- agent2Score.get(roundNr);
					if (diff > 0) {
						this.selectedAgent = agent1;
						setRockOrder(darkRockOrder);
					} else if (diff < 0) {
						this.selectedAgent = agent2;
						setRockOrder(lightRockOrder);
					}

					if (roundNr == numberOfRounds + extraRounds - 1) {
						int sumAgent1 = sum(agent1Score);
						int sumAgent2 = sum(agent2Score);
						if (sumAgent1 == sumAgent2) {
							extraRounds++;
						}
					}

					roundNr++;
					log.info(format(ROUND_COMPLETED, roundNr));
					try {
						setProgress(progress, 0, totalProgress);
					} catch (IllegalArgumentException e) {
						log.error("Set progress: " + progress + " "
								+ totalProgress, e);
					}
				} // end while (round < numberOfRounds)
				
				int agent1Sum,agent2Sum;
				if ((agent1Sum = sum(agent1Score)) > (agent2Sum = sum(agent2Score))) {
					agent1MatchWinnerCount++;
					
				} else {
					agent2MatchWinnerCount++;
					
				}
				
				agent1MatchPointsTotal += agent1Sum;
				agent2MatchPointsTotal += agent2Sum;
				
				teardown();
				match++;
			}

			simulator.setSimulatorReady(true);

		}catch (Exception e) {
			log.error(RUN_EXCEPTION, e);
			simulator.setSimulatorReady(true);
		}
		return null;
	}

	@Override
	protected void succeeded(Void ignored) {
		String ressultStatistics = format(LOG_MATCH,
				(agent1MatchWinnerCount + agent2MatchWinnerCount), agent1
						.getAgentName(), agent1MatchWinnerCount,
				agent2MatchWinnerCount, agent2.getAgentName(),
				agent1MatchPointsTotal, agent2MatchPointsTotal);

		setMessage(ressultStatistics);
		Toolkit.getDefaultToolkit().beep();
	}

	private TurnEvent turnEvent;

	private AfterTurnEvent afterEvent;

	/**
	 * 
	 * @param player
	 *            agent that controls broom.
	 * @param rockID
	 *            rock id controlled by broom.
	 * @param nextRockID
	 *            rock id after.
	 * @param rockCount
	 *            total rocks set in current round.
	 * @param roundCount
	 *            for current round.
	 * @param numberOfRounds
	 *            total number of rounds (so far).
	 * @throws InterruptedException
	 */
	private void playRock(Agent player, int rockID, int nextRockID, int rockCount,
			int roundCount, int numberOfRounds) throws InterruptedException {

		boolean hasHammer = ((player == agent1 && rockOrder == lightRockOrder) || (player == agent2 && rockOrder == darkRockOrder));

		turnEvent = new TurnEvent(simulator.getAgentControls(), rockCollection,
				simulator.getRocksInPlay(), player == agent1, rockID,
				roundCount + 1, numberOfRounds, hasHammer, getOpponent(player));

		player.onTurn(turnEvent);

		Point2D broomPos = null;
		int splitTime = 0;
		boolean isOutTurn = false;
		RockSender rockSender = simulator.getRockSender();
		
		// get input from human
		if (player.isInterActive()) {
			rockSender
					.setRock(rockID, nextRockID, rockCount, player.isInterActive());
			while (true) {
				Thread.sleep(INTERACTIVE_SLEEP);
				boolean isWaiting = rockSender.isWaiting();

				if (!isWaiting) {
					broomPos = simulator.getRockBroomControl().getHumanBroom();
					splitTime = simulator.getRockBroomControl().getHumanSplit();
					isOutTurn = simulator.getRockBroomControl().getHumanTurn();
					break;
				}
			}
		} else {
			BroomControls agentControls = simulator.getAgentControls();
			// keep the broom properties the agent set.
			broomPos = (Point2D) agentControls.getBroom().clone();
			splitTime = agentControls.getSplitTimeMillies().getValue();
			isOutTurn = agentControls.getOutTurn();

			// distort the broom properties
			simulator.getDistorter().distortBroom(simulator.getAgentControls());

			Thread.sleep(AGENT_SLEEP);
			rockSender.setRock(rockID, nextRockID, rockCount);
		}

		// if human, it sends null references...
		afterEvent = new AfterTurnEvent(simulator.getAgentControls(),
				rockCollection, simulator.getRocksInPlay(), player == agent1,
				rockID, roundCount + 1, numberOfRounds, broomPos, splitTime,
				isOutTurn, hasHammer, getOpponent(player));
		player.afterTurn(afterEvent);

	}

	/**
	 * Get next agent.
	 * 
	 * @param prev
	 *            agent.
	 * @return the next (other) agent.
	 */
	private Agent nextAfter(Agent prev) {
		return (prev == agent1) ? agent2 : agent1;
	}

	private String getOpponent(Agent agent) {
		return nextAfter(agent).getAgentName();
	}
	
	public void computeScore() {
		int score = 0;

		final RockSet<Pos> ipos = simulator.getRockCollection().getCurves()
				.getCurrentPos();

		List<RockComparable> sortedRocks = new ArrayList<RockComparable>();

		for (int i = 0; i < 16; i++) {
			RockComparable rock = new RockComparable(i, ipos.getRock(i));
			sortedRocks.add(rock);
		}

		Collections.sort((List<RockComparable>) sortedRocks);

		if (sortedRocks.get(0).isDark()) {
			int i = 0;
			while (sortedRocks.get(i).isDark() && sortedRocks.get(i).inHouse()
					&& i < sortedRocks.size() - 1) {
				score++;
				i++;
			}
			agent1Score.add(score);
			agent2Score.add(0);
		} else {
			int i = 0;
			while (!(sortedRocks.get(i).isDark())
					&& sortedRocks.get(i).inHouse() && i < sortedRocks.size()) {
				score++;
				i++;
			}
			agent1Score.add(0);
			agent2Score.add(score);
		}
	}

	private void notifyPlayers(ReadOnlyRockCollection rockPositions, int round) {

		boolean darkHasHammer = (rockOrder == lightRockOrder) ? true : false;

		int diff = agent1Score.get(round) - agent2Score.get(round);
		if (diff < 0) {
			int points = agent2Score.get(round);
			roundWin(agent2, rockPositions, round, points, !darkHasHammer);
			roundLoose(agent1, rockPositions, round, points, darkHasHammer);
		} else if (diff > 0) {
			int points = agent1Score.get(round);
			roundWin(agent1, rockPositions, round, points, darkHasHammer);
			roundLoose(agent2, rockPositions, round, points, !darkHasHammer);
		} else {
			int points = 0;
			roundTie(agent1, rockPositions, round, points, darkHasHammer);
			roundTie(agent2, rockPositions, round, points, !darkHasHammer);
		}
	}

	private boolean hasHammer(Agent agent, boolean darkHasHammer) {
		return (darkHasHammer && agent == agent1) || (!darkHasHammer && agent == agent2);
	}
	
	private void roundTie(Agent agent, ReadOnlyRockCollection rockPositions,
			int round, int points, boolean darkHasHammer) {
		agent.onTie(new RoundEndedEvent(rockPositions, simulator
				.getRocksInPlay(), agent == agent1, round, points, Result.TIE,
				hasHammer(agent, darkHasHammer)));
	}

	private void roundLoose(Agent agent, ReadOnlyRockCollection rockPositions,
			int round, int points, boolean darkHasHammer) {
		agent.onLoose(new RoundEndedEvent(rockPositions, simulator
				.getRocksInPlay(), agent == agent1, round, points, Result.WIN,
				hasHammer(agent, darkHasHammer)));
	}

	private void roundWin(Agent agent, ReadOnlyRockCollection rockPositions,
			int round, int points, boolean darkHasHammer) {
		agent.onWin(new RoundEndedEvent(rockPositions, simulator
				.getRocksInPlay(), agent == agent1, round, points, Result.WIN,
				hasHammer(agent, darkHasHammer)));
	}

	public void printRoundScore(int round, int s1, int s2) {
		Agent winner = (s1 > s2) ? agent1 : agent2;
		int winnerScore = Math.max(s1, s2);
		if (s1 == s2) {
			log.info(format(MATCH_TIE, (round + 1)));
		} else {
			log.info(format(MATCH_ENDED, (round + 1), winner.getAgentName(),
					winnerScore, (winner.equals(agent1) ? "red" : "yellow")));
		}
	}

	private int sum(List<Integer> score) {
		int totalScore = 0;
		for (int i : score) {
			totalScore += i;
		}
		return totalScore;
	}

	private void setRockOrder(int[] order) {
		this.rockOrder = order;
	}

	private int[] getRockOrder() {
		return rockOrder;
	}

	private String format(String format, Object... args) {
		return String.format(format, args);
	}

	private int[] rockOrder = darkRockOrder;

	private static int ROUND_END = RockBroomControl.ROUND_END;

	private static int[] lightRockOrder = { 1, 0, 3, 2, 5, 4, 7, 6, 9, 8, 11,
			10, 13, 12, 15, 14, ROUND_END };
	private static int[] darkRockOrder = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
			11, 12, 13, 14, 15, ROUND_END };

}
