package no.skippy.sim.control;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Iterator;
import java.util.Map.Entry;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import no.skippy.sim.model.RockCollection;
import no.skippy.sim.view.SimulatorView;

import org.apache.commons.logging.Log;
import org.jcurl.core.api.ComputedTrajectorySet;
import org.jcurl.core.api.Rock;
import org.jcurl.core.api.RockSet;
import org.jcurl.core.api.RockType.Pos;
import org.jcurl.core.log.JCLoggerFactory;
import org.jcurl.core.ui.ChangeManager;
import org.jcurl.math.R1RNFunction;
import org.jcurl.zui.piccolo.PTrajectoryFactory;

import edu.umd.cs.piccolo.PNode;

/**
 * Controller class.
 * Handles Zooming, and updating UI when {@link RockCollection} is changed.
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class SimulatorViewController implements HasChanger, ChangeListener, PropertyChangeListener {
	
	private static Log log = JCLoggerFactory.getLogger(SimulatorViewController.class);
	
	private SimulatorView view;
	private RockCollection rockCollection;
	
	private ChangeManager changer;
	PropertyChangeSupport changeSupport = new PropertyChangeSupport(this);
	private final PTrajectoryFactory trajectoryFactory = new PTrajectoryFactory.Fancy();
	
//	TODO: move to resource props?
	private static final int AnimationMillis = 0;
	
	public SimulatorViewController(SimulatorView view, RockCollection rockCollection) {
		this.view = view;
		this.rockCollection = rockCollection;

		if (view != null) {
			view.addPropertyChangeListener(this);
			view.getCanvas().addInputEventListener(new MouseWheelHandler(view));
		}
	}
	
	public ChangeManager getChanger() {
		return ChangeManager.getTrivial(changer);
	}

	public void setChanger(ChangeManager changer) {
		final ChangeManager old = this.changer;
		if (old == changer) {
			return;
		}
		changeSupport.firePropertyChange("changer", old, this.changer = changer);
	}

	@SuppressWarnings("unchecked")
	public void stateChanged(ChangeEvent event) {
		if (event.getSource() instanceof Rock) {
			RockCollection rocksModel = rockCollection;
			final Rock<Pos> r = (Rock<Pos>) event.getSource();
			final PNode n = rocksModel.getRocNodeMap().get(r);
			
			syncM2V(r, n, AnimationMillis);
			// update the trajectory path, current only.
			if (n == null) return;
			if (!Boolean.TRUE.equals(n
					.getAttribute(RockCollection.ATTR_TRIGGER_CURVE_UPDATE)))
				return;
			final int i16 = (Integer) n.getAttribute(RockCollection.ATTR_IDX16);
			syncM2V(rocksModel.getCurves(), i16, rocksModel
					.getPath(), trajectoryFactory);
		}
	}

	public void propertyChange(PropertyChangeEvent evt) {
//		System.out.println("property changed in view: REMOVE ME?");
	}
	
	// Static helpers.
	public PNode syncM2V(final RockSet<Pos> src, final int i16, final PNode dst) {
		final Rock<Pos> r = src.getRock(i16);
		syncM2V(r, dst, 0);
		dst.addAttribute(RockCollection.ATTR_ROCKSET, src);
		dst.addAttribute(RockCollection.ATTR_ROCK, r);
		dst.addAttribute(RockCollection.ATTR_IDX16, i16);
		rockCollection.getRocNodeMap().put(r, dst);
		return dst;
	}

	/** update one curve's path */
	public static void syncM2V(final ComputedTrajectorySet src, final int i16,
			final PNode[] dst, final PTrajectoryFactory tf) {
		syncM2V(src.getCurveStore().iterator(i16), i16, dst[i16], tf);
	}

	/** update one curves' path */
	private static void syncM2V(
			final Iterator<Entry<Double, R1RNFunction>> src, final int i16,
			final PNode dst, final PTrajectoryFactory tf) {
		if (log.isDebugEnabled())
			log.debug("path " + i16);
		tf.refresh(src, RockSet.isDark(i16), 0, 30, dst);
	}

	/** update one rock */
	public static void syncM2V(final Rock<Pos> src, final PNode dst,
			final long dt) {
		if (src == null || dst == null) {
			if (log.isDebugEnabled())
				log.debug(src + " " + dst);
			return;
		}
		if (log.isDebugEnabled())
			log.debug("rock " + dst.getAttribute(RockCollection.ATTR_IDX16));
		if (dt >= 0)
			dst.animateToTransform(src.getAffineTransform(), dt);
		else
			dst.setTransform(src.getAffineTransform());
	}
	
	public void setCurves(ComputedTrajectorySet curves) {
		if (rockCollection.getCurves() != null) {
			removeCL(rockCollection.getCurves().getInitialPos());
			removeCL(rockCollection.getCurves().getCurrentPos());
		}
		rockCollection.setCurves(curves);
		rockCollection.getRocNodeMap().clear();
		
		if (rockCollection.getCurves() != null) {
			final RockSet<Pos> ip = curves.getInitialPos();
			final RockSet<Pos> cp = curves.getCurrentPos();
			
			addCL(ip);
			addCL(cp);
			for (int i16 = RockSet.ROCKS_PER_SET - 1; i16 >= 0; i16--) {
				syncM2V(ip, i16, rockCollection.getInitialRocks()[i16]);
				syncM2V(cp, i16, rockCollection.getCurrentRocks()[i16]);
				syncM2V(curves, i16, rockCollection.getPath(), trajectoryFactory );
			}
		}
		if (view != null)
		view.getBroomModel().init(curves);
		rockCollection.getRocks().setVisible(curves != null);
	}
	
	protected void addCL(final RockSet<Pos> s) {
		if (s == null)
			return;
		for (int i16 = RockSet.ROCKS_PER_SET - 1; i16 >= 0; i16--)
			s.getRock(i16).addChangeListener(this);
	}

	protected void removeCL(final RockSet<Pos> s) {
		if (s == null)
			return;
		for (int i16 = RockSet.ROCKS_PER_SET - 1; i16 >= 0; i16--)
			s.getRock(i16).removeChangeListener(this);
	}
}
