package no.skippy.sim.view;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.GridLayout;

import javax.swing.ActionMap;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;

import no.skippy.agent.Agent;
import no.skippy.agent.BasicAgent;

import org.apache.commons.logging.Log;
import org.jcurl.core.log.JCLoggerFactory;
import org.jdesktop.application.Action;
import org.jdesktop.application.ApplicationContext;
import org.jdesktop.application.ResourceMap;

/**
 * NewSimulatorDialog. Select agents, etc, and start the simulator.
 * 
 * @author <a href="mailto:frode.aa@gmail.com">Frode Aannevik </a>
 */
public class NewSimulatorDialog extends JDialog {

	/** */
	private static final long serialVersionUID = 1L;

	public static final int ERROR_OPTION = -1;
	public static final int CANCEL_OPTION = 1;

	private DefaultListModel agent1, agent2;
	private JList list1, list2;
	private ApplicationContext ctx;

	private JTextField numOfRoundsField;
	private JTextField repeatCountField;
	private JCheckBox exitOnComplete;

	private JFrame parent;

	private static Log logger = JCLoggerFactory
			.getLogger(NewSimulatorDialog.class);

	private boolean runSimulatorApproved = false;

	public NewSimulatorDialog(ApplicationContext ctx, JFrame parent,
			Agent[] agents) {
		super(parent, ctx.getResourceMap(NewSimulatorDialog.class).getString(
				"DialogTitle.text"), true);

		this.ctx = ctx;
		this.parent = parent;

		agent1 = new DefaultListModel();
		agent2 = new DefaultListModel();

		for (int i= 0; i < agents.length; i++) {
			agent1.add(i, agents[i]);
			agent2.add(i, agents[i]);
		}

		initComponents();
		setResizable(false);
	}

	public NewSimulatorDialog showDialog(final JFrame parent) {
		pack();
		setLocationRelativeTo(parent);
		setVisible(true);
		return this;
	}

	private boolean validateForm() {
		ResourceMap rs = ctx.getResourceMap(NewSimulatorDialog.class);

		String agentNotSelectedError = rs.getString("error1.text");
		if (list1.getSelectedValue() == null
				|| list2.getSelectedValue() == null) {
			showErrorDialog(agentNotSelectedError, new NullPointerException());
			return false;
		}

		String numOfRoundsError = rs.getString("error2.text");
		if (numOfRoundsField.getText().length() == 0) {
			showErrorDialog(numOfRoundsError, new NullPointerException());
			return false;
		}

		String numOfMatchError = rs.getString("error3.text");
		if (repeatCountField.getText().length() == 0) {
			showErrorDialog(numOfMatchError, new NullPointerException());
			return false;
		}

		String parseIntError = rs.getString("error4.text");
		try {
			Integer.parseInt(numOfRoundsField.getText().trim());
		} catch (NumberFormatException e) {
			showErrorDialog(String.format(parseIntError, numOfRoundsField
					.getText()), e);
			return false;
		}

		try {
			Integer.parseInt(repeatCountField.getText().trim());
		} catch (NumberFormatException e) {
			showErrorDialog(String.format(parseIntError, repeatCountField
					.getText()), e);
			return false;
		}

		return true;
	}

	@Action
	public void abortAction() {
		logger.info("Simulator aborted.");
		runSimulatorApproved = false;
		setVisible(false);
	}

	@Action
	public void okAction() {
		if (!validateForm()) {
			logger.warn("error: check dialog inputs");
			return;
		}
		runSimulatorApproved = true;
		logger.info("Simulator starting.");
		setVisible(false);
	}

	public int getNumberOfRounds() {
		String value = numOfRoundsField.getText();
		return Integer.parseInt(value);
	}

	public int getNumberOfMatch() {
		return Integer.parseInt(repeatCountField.getText());
	}

	public boolean isExitOnCompleteSelected() {
		return exitOnComplete.isSelected();
	}
	
	public Class<?> getAgent1() {
		return getAgent(list1);
	}

	public Class<?> getAgent2() {
		return getAgent(list2);
	}

	public boolean isRunSimulatorApproved() {
		return runSimulatorApproved;
	}

	private void initComponents() {
		setLayout(new BorderLayout());

		JPanel agentView = new JPanel(new GridLayout(1, 2));
		agentView.setBackground(Color.WHITE);
		agentView.add(createAgentList(list1 = new JList(agent1), "Agent 1"));
		agentView.add(createAgentList(list2 = new JList(agent2), "Agent 2"));

		list1.setSelectedIndex(Math.max(0, agent1.getSize() - 1));
		list2.setSelectedIndex(Math.max(0, agent2.getSize() - 4));

		JPanel cc = new JPanel();
		cc.add(createControl(numOfRoundsField = new JTextField(),
				"Number of rounds:", 10));
		cc.add(createControl(repeatCountField = new JTextField(),
				"Number of match:", 1));

		JPanel buttons = new JPanel();
		buttons.setLayout(new BoxLayout(buttons, BoxLayout.Y_AXIS));
		buttons.add(Box.createVerticalStrut(7));
		buttons.add(createButtons(new String[] { "abortAction", "okAction" }));

		JPanel controls = new JPanel(new GridLayout(1, 2));
		controls.add(cc);
		controls.add(buttons);

		add(Box.createVerticalStrut(20), BorderLayout.NORTH);
		add(agentView, BorderLayout.CENTER);
		add(controls, BorderLayout.SOUTH);

	}

	private JPanel createButtons(String[] actionNames) {
		JPanel v = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		v.setBackground(getBackground());
		v.add(exitOnComplete = new JCheckBox("Exit on complete?"));
		ActionMap amap = ctx.getActionMap(this);
		for (String actionName : actionNames) {
			JButton abutton = new JButton(amap.get(actionName));
			abutton.setName(actionName);
			v.add(abutton);
		}

		return v;
	}

	private JScrollPane createAgentList(JList agents, String borderTitle) {
		JScrollPane scrollPane = new JScrollPane();
		scrollPane.setViewportView(agents);
		agents.setCellRenderer(new AgentCellRender());
		scrollPane.setBorder(BorderFactory.createTitledBorder(borderTitle));
		return scrollPane;
	}

	private JPanel createControl(JTextField field, String lblText,
			int defaultValue) {
		JPanel panel = new JPanel(new FlowLayout(FlowLayout.LEFT));

		field.setColumns(3);
		field.setText("" + defaultValue);
		field.setHorizontalAlignment(JTextField.RIGHT);

		JLabel label = new JLabel(lblText);
		label.setLabelFor(field);

		panel.add(label);
		panel.add(field);

		return panel;
	}

	private Class<?> getAgent(JList list) {
		Object aa = list.getSelectedValue();
		BasicAgent agentclass = null;
		if (aa instanceof BasicAgent) {
			agentclass = (BasicAgent) aa;
			return agentclass.getClass();
		}
		throw new UnsupportedOperationException(
				"Wrong agent class type, all agent must be of BasicAgent");
	}

	private void showErrorDialog(final String message, final Exception e) {
		JOptionPane.showMessageDialog(parent, "Error: " + message, "Error",
				JOptionPane.ERROR_MESSAGE);
		logger.warn("Error " + message, e);
	}

	class AgentCellRender extends JLabel implements ListCellRenderer {
		/** */
		private static final long serialVersionUID = 1L;

		public AgentCellRender() {
			setOpaque(true); // don't paint behind the component.
		}

		public Component getListCellRendererComponent(JList list, Object value,
				int index, boolean isSelected, boolean cellHasFocus) {
			setText(((Agent) value).getAgentName()
					+ (((BasicAgent) value).isInterActive() ? "   (Human)" : ""));
			if (isSelected) {
				setBackground(list.getSelectionBackground());
			} else {
				setBackground(list.getBackground());
			}
			return this;
		}
	}

}